from django.test import SimpleTestCase

from evennia import DefaultChannel
from evennia.commands.default.comms import CmdChannel
from evennia.utils.create import create_message
from evennia.utils.test_resources import BaseEvenniaTest


class TestCommsNickMatchesCommand(SimpleTestCase):
    def test(self):
        """
        Verifies that the nick being set by DefaultChannel matches the channel
        command key.
        """
        self.assertTrue(DefaultChannel.channel_msg_nick_replacement.startswith(CmdChannel.key))


class ObjectCreationTest(BaseEvenniaTest):
    def test_channel_create(self):
        description = "A place to talk about coffee."

        obj, errors = DefaultChannel.create("coffeetalk", description=description)
        self.assertTrue(obj, errors)
        self.assertFalse(errors, errors)
        self.assertEqual(description, obj.db.desc)

    def test_message_create(self):
        msg = create_message("peewee herman", "heh-heh!", header="mail time!")
        self.assertTrue(msg)
        self.assertEqual(str(msg), "peewee herman->: heh-heh!")


class ChannelSubscriptionTests(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.default_channel, _ = DefaultChannel.create(
            "catlovers", description="A place for feline fanciers."
        )
        self.default_channel.connect(self.obj1)

    def test_subscribe_unsubscribe(self):
        self.default_channel.connect(self.char1)
        self.assertTrue(self.default_channel.subscriptions.has(self.char1))
        self.assertEqual(
            self.char1.nicks.nickreplace("catlovers I love cats!"),
            "@channel catlovers = I love cats!",
        )
        self.default_channel.disconnect(self.char1)
        self.assertFalse(self.default_channel.subscriptions.has(self.char1))
        self.assertEqual(
            self.char1.nicks.nickreplace("catlovers I love cats!"),
            "catlovers I love cats!",
        )


class ChannelWholistTests(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.default_channel, _ = DefaultChannel.create(
            "coffeetalk", description="A place to talk about coffee."
        )
        self.default_channel.connect(self.obj1)

    def test_wholist_shows_subscribed_objects(self):
        expected = "Obj"
        result = self.default_channel.wholist
        self.assertEqual(expected, result)

    def test_wholist_shows_none_when_empty(self):
        # No one hates dogs
        empty_channel, _ = DefaultChannel.create(
            "doghaters", description="A place where dog haters unite."
        )
        expected = "<None>"
        result = empty_channel.wholist
        self.assertEqual(expected, result)

    def test_wholist_does_not_show_muted_objects(self):
        self.default_channel.mute(self.obj2)
        expected = "Obj"
        result = self.default_channel.wholist
        self.assertEqual(expected, result)

    def test_wholist_shows_connected_object_as_bold(self):
        self.default_channel.connect(self.char1)
        expected = "Obj, |wChar|n"
        result = self.default_channel.wholist
        self.assertEqual(expected, result)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-07-05 17:36


from django.db import migrations


def forwards(apps, schema_editor):

    try:
        apps.get_model("accounts", "AccountDB")
    except LookupError:
        return
    AccountDB = apps.get_model("accounts", "AccountDB")
    ScriptDB = apps.get_model("scripts", "ScriptDB")

    for script in ScriptDB.objects.all():
        account = script.db_account
        if account:
            account = AccountDB.objects.get(id=account.id)
            script.db_account = account
            script.save(update_fields=["db_account"])


class Migration(migrations.Migration):

    dependencies = [("scripts", "0009_scriptdb_db_account")]

    operations = [migrations.RunPython(forwards)]



"""
This sub-package defines the help system of Evennia. It is pretty
simple, mainly consisting of a database model to hold help entries.
The auto-cmd-help is rather handled by the default 'help' command
itself.

"""



# -*- coding: utf-8 -*-
"""
This sub-package defines the lock (access) mechanism of Evennia. All
lock strings are processed through the lockhandler in this package. It
also contains the default lock functions used in lock definitions.

"""



# -*- coding: utf-8 -*-


import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("objects", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("typeclasses", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="objectdb",
            name="db_account",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.SET_NULL,
                verbose_name="account",
                to=settings.AUTH_USER_MODEL,
                help_text="an Account connected to this object, if any.",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="objectdb",
            name="db_tags",
            field=models.ManyToManyField(
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
                null=True,
            ),
            preserve_default=True,
        ),
    ]



"""
Test tutorial_world/mob

"""

from mock import patch
from twisted.internet.base import DelayedCall
from twisted.trial.unittest import TestCase as TwistedTestCase

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object
from evennia.utils.test_resources import BaseEvenniaTest, mockdeferLater, mockdelay

from . import mob
from . import objects as tutobjects
from . import rooms as tutrooms


class TestTutorialWorldMob(BaseEvenniaTest):
    def test_mob(self):
        mobobj = create_object(mob.Mob, key="mob")
        self.assertEqual(mobobj.db.is_dead, True)
        mobobj.set_alive()
        self.assertEqual(mobobj.db.is_dead, False)
        mobobj.set_dead()
        self.assertEqual(mobobj.db.is_dead, True)
        mobobj._set_ticker(0, "foo", stop=True)
        # TODO should be expanded with further tests of the modes and damage etc.


#  test tutorial_world/objects


DelayedCall.debug = True


class TestTutorialWorldObjects(TwistedTestCase, BaseEvenniaCommandTest):
    def tearDown(self):
        self.char1.delete()
        super(BaseEvenniaCommandTest, self).tearDown()

    def test_tutorialobj(self):
        obj1 = create_object(tutobjects.TutorialObject, key="tutobj")
        obj1.reset()
        self.assertEqual(obj1.location, obj1.home)

    def test_readable(self):
        readable = create_object(tutobjects.TutorialReadable, key="book", location=self.room1)
        readable.db.readable_text = "Text to read"
        self.call(tutobjects.CmdRead(), "book", "You read book:\n  Text to read", obj=readable)

    def test_climbable(self):
        climbable = create_object(tutobjects.TutorialClimbable, key="tree", location=self.room1)
        self.call(
            tutobjects.CmdClimb(),
            "tree",
            "You climb tree. Having looked around, you climb down again.",
            obj=climbable,
        )
        self.assertEqual(
            self.char1.tags.get("tutorial_climbed_tree", category="tutorial_world"),
            "tutorial_climbed_tree",
        )

    def test_obelisk(self):
        obelisk = create_object(tutobjects.Obelisk, key="obelisk", location=self.room1)
        self.assertEqual(obelisk.return_appearance(self.char1).startswith("|cobelisk("), True)

    @patch("evennia.contrib.tutorials.tutorial_world.objects.delay", mockdelay)
    @patch("evennia.scripts.taskhandler.deferLater", mockdeferLater)
    def test_lightsource(self):
        light = create_object(tutobjects.LightSource, key="torch", location=self.room1)
        self.call(
            tutobjects.CmdLight(),
            "",
            "A torch on the floor flickers and dies.|You light torch.",
            obj=light,
        )
        self.assertFalse(light.pk)

    @patch("evennia.contrib.tutorials.tutorial_world.objects.delay", mockdelay)
    @patch("evennia.scripts.taskhandler.deferLater", mockdeferLater)
    def test_crumblingwall(self):
        wall = create_object(tutobjects.CrumblingWall, key="wall", location=self.room1)
        wall.db.destination = self.room2.dbref
        self.assertFalse(wall.db.button_exposed)
        self.assertFalse(wall.db.exit_open)
        wall.db.root_pos = {"yellow": 0, "green": 0, "red": 0, "blue": 0}
        self.call(
            tutobjects.CmdShiftRoot(),
            "blue root right",
            "You shove the root adorned with small blue flowers to the right.",
            obj=wall,
        )
        self.call(
            tutobjects.CmdShiftRoot(),
            "red root left",
            "You shift the reddish root to the left.",
            obj=wall,
        )
        self.call(
            tutobjects.CmdShiftRoot(),
            "yellow root down",
            "You shove the root adorned with small yellow flowers downwards.",
            obj=wall,
        )
        self.call(
            tutobjects.CmdShiftRoot(),
            "green root up",
            (
                "You shift the weedy green root upwards.|Holding aside the root you "
                "think you notice something behind it ..."
            ),
            obj=wall,
        )
        self.call(
            tutobjects.CmdPressButton(),
            "",
            (
                "You move your fingers over the suspicious depression, then gives it a "
                "decisive push. First"
            ),
            obj=wall,
        )
        # we patch out the delay, so these are closed immediately
        self.assertFalse(wall.db.button_exposed)
        self.assertFalse(wall.db.exit_open)

    def test_weapon(self):
        weapon = create_object(tutobjects.TutorialWeapon, key="sword", location=self.char1)
        self.call(
            tutobjects.CmdAttack(), "Char", "You stab with sword.", obj=weapon, cmdstring="stab"
        )
        self.call(
            tutobjects.CmdAttack(), "Char", "You slash with sword.", obj=weapon, cmdstring="slash"
        )

    def test_weaponrack(self):
        rack = create_object(tutobjects.TutorialWeaponRack, key="rack", location=self.room1)
        rack.db.available_weapons = ["sword"]
        self.call(tutobjects.CmdGetWeapon(), "", "You find Rusty sword.", obj=rack)


class TestTutorialWorldRooms(BaseEvenniaCommandTest):
    def test_cmdtutorial(self):
        room = create_object(tutrooms.TutorialRoom, key="tutroom")
        self.char1.location = room
        self.call(tutrooms.CmdTutorial(), "", "Sorry, there is no tutorial help available here.")
        self.call(
            tutrooms.CmdTutorialSetDetail(),
            "detail;foo;foo2 = A detail",
            "Detail set: 'detail;foo;foo2': 'A detail'",
            obj=room,
        )
        self.call(tutrooms.CmdTutorialLook(), "", "tutroom(", obj=room)
        self.call(tutrooms.CmdTutorialLook(), "detail", "A detail", obj=room)
        self.call(tutrooms.CmdTutorialLook(), "foo", "A detail", obj=room)
        room.delete()

    def test_weatherroom(self):
        room = create_object(tutrooms.WeatherRoom, key="weatherroom")
        room.update_weather()
        tutrooms.TICKER_HANDLER.remove(
            interval=room.db.interval, callback=room.update_weather, idstring="tutorial"
        )
        room.delete()

    def test_introroom(self):
        room = create_object(tutrooms.IntroRoom, key="introroom")
        room.at_object_receive(self.char1, self.room1)

    def test_bridgeroom(self):
        room = create_object(tutrooms.BridgeRoom, key="bridgeroom")
        room.update_weather()
        self.char1.move_to(room, move_type="teleport")
        self.call(
            tutrooms.CmdBridgeHelp(),
            "",
            "You are trying hard not to fall off the bridge ...",
            obj=room,
        )
        self.call(
            tutrooms.CmdLookBridge(),
            "",
            "bridgeroom\nYou are standing very close to the the bridge's western foundation.",
            obj=room,
        )
        room.at_object_leave(self.char1, self.room1)
        tutrooms.TICKER_HANDLER.remove(
            interval=room.db.interval, callback=room.update_weather, idstring="tutorial"
        )
        room.delete()

    def test_darkroom(self):
        room = create_object(tutrooms.DarkRoom, key="darkroom")
        self.char1.move_to(room, move_type="teleport")
        self.call(tutrooms.CmdDarkHelp(), "", "Can't help you until")

    def test_teleportroom(self):
        create_object(tutrooms.TeleportRoom, key="teleportroom")

    def test_outroroom(self):
        create_object(tutrooms.OutroRoom, key="outroroom")



"""
TickerHandler

This implements an efficient Ticker which uses a subscription
model to 'tick' subscribed objects at regular intervals.

The ticker mechanism is used by importing and accessing
the instantiated TICKER_HANDLER instance in this module. This
instance is run by the server; it will save its status across
server reloads and be started automaticall on boot.

Example:

```python
    from evennia.scripts.tickerhandler import TICKER_HANDLER

    # call tick myobj.at_tick(*args, **kwargs) every 15 seconds
    TICKER_HANDLER.add(15, myobj.at_tick, *args, **kwargs)
```

You supply the interval to tick and a callable to call regularly with
any extra args/kwargs. The callable should either be a stand-alone
function in a module *or* the method on a *typeclassed* entity (that
is, on an object that can be safely and stably returned from the
database).  Functions that are dynamically created or sits on
in-memory objects cannot be used by the tickerhandler (there is no way
to reference them safely across reboots and saves).

The handler will transparently set
up and add new timers behind the scenes to tick at given intervals,
using a TickerPool - all callables with the same interval will share
the interval ticker.

To remove:

```python
    TICKER_HANDLER.remove(15, myobj.at_tick)
```

Both interval and callable must be given since a single object can be subscribed
to many different tickers at the same time. You can also supply `idstring`
as an identifying string if you ever want to tick the callable at the same interval
but with different arguments (args/kwargs are not used for identifying the ticker). There
is also `persistent=False` if you don't want to make a ticker that don't survive a reload.
If either or both `idstring` or `persistent` has been changed from their defaults, they
must be supplied to the `TICKER_HANDLER.remove` call to properly identify the ticker
to remove.

The TickerHandler's functionality can be overloaded by modifying the
Ticker class and then changing TickerPool and TickerHandler to use the
custom classes

```python
class MyTicker(Ticker):
    # [doing custom stuff]

class MyTickerPool(TickerPool):
    ticker_class = MyTicker
class MyTickerHandler(TickerHandler):
    ticker_pool_class = MyTickerPool
```

If one wants to duplicate TICKER_HANDLER's auto-saving feature in
a  custom handler one can make a custom `AT_STARTSTOP_MODULE` entry to
call the handler's `save()` and `restore()` methods when the server reboots.

"""
import inspect

from django.core.exceptions import ObjectDoesNotExist
from twisted.internet.defer import inlineCallbacks

from evennia.scripts.scripts import ExtendedLoopingCall
from evennia.server.models import ServerConfig
from evennia.utils import inherits_from, variable_from_module
from evennia.utils.dbserialize import dbserialize, dbunserialize, pack_dbobj
from evennia.utils.logger import log_err, log_trace

_GA = object.__getattribute__
_SA = object.__setattr__


_ERROR_ADD_TICKER = """TickerHandler: Tried to add an invalid ticker:
{store_key}
Ticker was not added."""

_ERROR_ADD_TICKER_SUB_SECOND = """You are trying to add a ticker running faster
than once per second. This is not supported and also probably not useful:
Spamming messages to the user faster than once per second serves no purpose in
a text-game, and if you want to update some property, consider doing so
on-demand rather than using a ticker.
"""


class Ticker(object):
    """
    Represents a repeatedly running task that calls
    hooks repeatedly. Overload `_callback` to change the
    way it operates.
    """

    @inlineCallbacks
    def _callback(self):
        """
        This will be called repeatedly every `self.interval` seconds.
        `self.subscriptions` contain tuples of (obj, args, kwargs) for
        each subscribing object.

        If overloading, this callback is expected to handle all
        subscriptions when it is triggered. It should not return
        anything and should not traceback on poorly designed hooks.
        The callback should ideally work under @inlineCallbacks so it
        can yield appropriately.

        The _hook_key, which is passed down through the handler via
        kwargs is used here to identify which hook method to call.

        """
        self._to_add = []
        self._to_remove = []
        self._is_ticking = True
        for store_key, (args, kwargs) in self.subscriptions.items():
            callback = yield kwargs.pop("_callback", "at_tick")
            obj = yield kwargs.pop("_obj", None)
            try:
                if callable(callback):
                    # call directly
                    yield callback(*args, **kwargs)
                    continue
                # try object method
                if not obj or not obj.pk:
                    # object was deleted between calls
                    self._to_remove.append(store_key)
                    continue
                else:
                    yield _GA(obj, callback)(*args, **kwargs)
            except ObjectDoesNotExist:
                log_trace("Removing ticker.")
                self._to_remove.append(store_key)
            except Exception:
                log_trace()
            finally:
                # make sure to re-store
                kwargs["_callback"] = callback
                kwargs["_obj"] = obj
        # cleanup - we do this here to avoid changing the subscription dict while it loops
        self._is_ticking = False
        for store_key in self._to_remove:
            self.remove(store_key)
        for store_key, (args, kwargs) in self._to_add:
            self.add(store_key, *args, **kwargs)
        self._to_remove = []
        self._to_add = []

    def __init__(self, interval):
        """
        Set up the ticker

        Args:
            interval (int): The stepping interval.

        """
        self.interval = interval
        self.subscriptions = {}
        self._is_ticking = False
        self._to_remove = []
        self._to_add = []
        # set up a twisted asynchronous repeat call
        self.task = ExtendedLoopingCall(self._callback)

    def validate(self, start_delay=None):
        """
        Start/stop the task depending on how many subscribers we have
        using it.

        Args:
            start_delay (int, optional): Time to way before starting.

        """
        subs = self.subscriptions
        if self.task.running:
            if not subs:
                self.task.stop()
        elif subs:
            self.task.start(self.interval, now=False, start_delay=start_delay)

    def add(self, store_key, *args, **kwargs):
        """
        Sign up a subscriber to this ticker.
        Args:
            store_key (str): Unique storage hash for this ticker subscription.
            args (any, optional): Arguments to call the hook method with.

        Keyword Args:
            _start_delay (int): If set, this will be
                used to delay the start of the trigger instead of
                `interval`.

        """
        if self._is_ticking:
            # protects the subscription dict from
            # updating while it is looping
            self._to_add.append((store_key, (args, kwargs)))
        else:
            start_delay = kwargs.pop("_start_delay", None)
            self.subscriptions[store_key] = (args, kwargs)
            self.validate(start_delay=start_delay)

    def remove(self, store_key):
        """
        Unsubscribe object from this ticker

        Args:
            store_key (str): Unique store key.

        """
        if self._is_ticking:
            # this protects the subscription dict from
            # updating while it is looping
            self._to_remove.append(store_key)
        else:
            self.subscriptions.pop(store_key, False)
            self.validate()

    def stop(self):
        """
        Kill the Task, regardless of subscriptions.

        """
        self.subscriptions = {}
        self.validate()


class TickerPool(object):
    """
    This maintains a pool of
    `evennia.scripts.scripts.ExtendedLoopingCall` tasks for calling
    subscribed objects at given times.

    """

    ticker_class = Ticker

    def __init__(self):
        """
        Initialize the pool.

        """
        self.tickers = {}

    def add(self, store_key, *args, **kwargs):
        """
        Add new ticker subscriber.

        Args:
            store_key (str): Unique storage hash.
            args (any, optional): Arguments to send to the hook method.

        """
        _, _, _, interval, _, _ = store_key
        if not interval:
            log_err(_ERROR_ADD_TICKER.format(store_key=store_key))
            return

        if interval not in self.tickers:
            self.tickers[interval] = self.ticker_class(interval)
        self.tickers[interval].add(store_key, *args, **kwargs)

    def remove(self, store_key):
        """
        Remove subscription from pool.

        Args:
            store_key (str): Unique storage hash to remove

        """
        _, _, _, interval, _, _ = store_key
        if interval in self.tickers:
            self.tickers[interval].remove(store_key)
            if not self.tickers[interval]:
                del self.tickers[interval]

    def stop(self, interval=None):
        """
        Stop all scripts in pool. This is done at server reload since
        restoring the pool will automatically re-populate the pool.

        Args:
            interval (int, optional): Only stop tickers with this interval.

        """
        if interval and interval in self.tickers:
            self.tickers[interval].stop()
        else:
            for ticker in self.tickers.values():
                ticker.stop()


class TickerHandler(object):
    """
    The Tickerhandler maintains a pool of tasks for subscribing
    objects to various tick rates.  The pool maintains creation
    instructions and and re-applies them at a server restart.

    """

    ticker_pool_class = TickerPool

    def __init__(self, save_name="ticker_storage"):
        """
        Initialize handler

        save_name (str, optional): The name of the ServerConfig
            instance to store the handler state persistently.

        """
        self.ticker_storage = {}
        self.save_name = save_name
        self.ticker_pool = self.ticker_pool_class()

    def _get_callback(self, callback):
        """
        Analyze callback and determine its consituents

        Args:
            callback (function or method): This is either a stand-alone
                function or class method on a typeclassed entitye (that is,
                an entity that can be saved to the database).
        Returns:
            ret (tuple): This is a tuple of the form `(obj, path, callfunc)`,
                where `obj` is the database object the callback is defined on
                if it's a method (otherwise `None`) and vice-versa, `path` is
                the python-path to the stand-alone function (`None` if a method).
                The `callfunc` is either the name of the method to call or the
                callable function object itself.
        Raises:
            TypeError: If the callback is of an unsupported type.

        """
        outobj, outpath, outcallfunc = None, None, None
        if callable(callback):
            if inspect.ismethod(callback):
                outobj = callback.__self__
                outcallfunc = callback.__func__.__name__
            elif inspect.isfunction(callback):
                outpath = "%s.%s" % (callback.__module__, callback.__name__)
                outcallfunc = callback
            else:
                raise TypeError(f"{callback} is not a method or function.")
        else:
            raise TypeError(f"{callback} is not a callable function or method.")

        if outobj and not inherits_from(outobj, "evennia.typeclasses.models.TypedObject"):
            raise TypeError(
                f"{callback} is a method on a normal object - it must "
                "be either a method on a typeclass, or a stand-alone function."
            )

        return outobj, outpath, outcallfunc

    def _store_key(self, obj, path, interval, callfunc, idstring="", persistent=True):
        """
        Tries to create a store_key for the object.

        Args:
            obj (Object, tuple or None): Subscribing object if any. If a tuple, this is
                a packed_obj tuple from dbserialize.
            path (str or None): Python-path to callable, if any.
            interval (int): Ticker interval. Floats will be converted to
                nearest lower integer value.
            callfunc (callable or str): This is either the callable function or
                the name of the method to call. Note that the callable is never
                stored in the key; that is uniquely identified with the python-path.
            idstring (str, optional): Additional separator between
                different subscription types.
            persistent (bool, optional): If this ticker should survive a system
                shutdown or not.

        Returns:
            store_key (tuple): A tuple `(packed_obj, methodname, outpath, interval,
                idstring, persistent)` that uniquely identifies the
                ticker. Here, `packed_obj` is the unique string representation of the
                object or `None`. The `methodname` is the string name of the method on
                `packed_obj` to call, or `None` if `packed_obj` is unset. `path` is
                the Python-path to a non-method callable, or `None`. Finally, `interval`
                `idstring` and `persistent` are integers, strings and bools respectively.

        """
        if interval < 1:
            raise RuntimeError(_ERROR_ADD_TICKER_SUB_SECOND)

        interval = int(interval)
        persistent = bool(persistent)
        packed_obj = pack_dbobj(obj)
        methodname = callfunc if callfunc and isinstance(callfunc, str) else None
        outpath = path if path and isinstance(path, str) else None
        return (packed_obj, methodname, outpath, interval, idstring, persistent)

    def save(self):
        """
        Save ticker_storage as a serialized string into a temporary
        ServerConf field. Whereas saving is done on the fly, if called
        by server when it shuts down, the current timer of each ticker
        will be saved so it can start over from that point.

        """
        if self.ticker_storage:
            # get the current times so the tickers can be restarted with a delay later
            start_delays = dict(
                (interval, ticker.task.next_call_time())
                for interval, ticker in self.ticker_pool.tickers.items()
            )

            to_save = {}

            # remove any subscription that lost its object and update the timers for the tickers
            for store_key, (args, kwargs) in self.ticker_storage.items():
                # unpack the store_key to reference its parts
                packedobj, callfunc, path, interval, idstring, persistent = store_key
                # verify that there's a valid obj+method or function path
                if (
                    callfunc
                    and ("_obj" in kwargs and kwargs["_obj"].pk)
                    and hasattr(kwargs["_obj"], callfunc)
                ) or path:
                    # this is a mutable, so it's updated in-place in ticker_storage
                    kwargs["_start_delay"] = start_delays.get(interval, None)
                    to_save[store_key] = (args, kwargs)
            ServerConfig.objects.conf(key=self.save_name, value=dbserialize(to_save))
        else:
            # make sure we have nothing lingering in the database
            ServerConfig.objects.conf(key=self.save_name, delete=True)

    def restore(self, server_reload=True):
        """
        Restore ticker_storage from database and re-initialize the
        handler from storage. This is triggered by the server at
        restart.

        Args:
            server_reload (bool, optional): If this is False, it means
                the server went through a cold reboot and all
                non-persistent tickers must be killed.

        """
        # load stored command instructions and use them to re-initialize handler
        restored_tickers = ServerConfig.objects.conf(key=self.save_name)
        if restored_tickers:
            # the dbunserialize will convert all serialized dbobjs to real objects

            restored_tickers = dbunserialize(restored_tickers)
            self.ticker_storage = {}
            for store_key, (args, kwargs) in restored_tickers.items():
                try:
                    # at this point obj is the actual object (or None) due to how
                    # the dbunserialize works
                    obj, callfunc, path, interval, idstring, persistent = store_key
                    if not persistent and not server_reload:
                        # this ticker will not be restarted
                        continue
                    if isinstance(callfunc, str) and not obj:
                        # methods must have an existing object
                        continue
                    # we must rebuild the store_key here since obj must not be
                    # stored as the object itself for the store_key to be hashable.
                    store_key = self._store_key(obj, path, interval, callfunc, idstring, persistent)

                    if obj and callfunc:
                        kwargs["_callback"] = callfunc
                        kwargs["_obj"] = obj
                    elif path:
                        modname, varname = path.rsplit(".", 1)
                        callback = variable_from_module(modname, varname)
                        kwargs["_callback"] = callback
                        kwargs["_obj"] = None
                    else:
                        # Neither object nor path - discard this ticker
                        log_err("Tickerhandler: Removing malformed ticker: %s" % str(store_key))
                        continue
                except Exception:
                    # this suggests a malformed save or missing objects
                    log_trace("Tickerhandler: Removing malformed ticker: %s" % str(store_key))
                    continue
                # if we get here we should create a new ticker
                self.ticker_storage[store_key] = (args, kwargs)
                self.ticker_pool.add(store_key, *args, **kwargs)

    def add(self, interval=60, callback=None, idstring="", persistent=True, *args, **kwargs):
        """
        Add subscription to tickerhandler

        Args:

            *args: Will be passed into the callback every time it's called. This must be
                data possible to pickle.

        Keyword Args:
            interval (int): Interval in seconds between calling
                `callable(*args, **kwargs)`
            callable (callable function or method): This
                should either be a stand-alone function or a method on a
                typeclassed entity (that is, one that can be saved to the
                database).
            idstring (str): Identifier for separating
                this ticker-subscription from others with the same
                interval. Allows for managing multiple calls with
                the same time interval and callback.
            persistent (bool): A ticker will always survive
                a server reload. If this is unset, the ticker will be
                deleted by a server shutdown.
            **kwargs Will be passed into the callback every time it is called.
                This must be data possible to pickle.

        Returns:
            store_key (tuple): The immutable store-key for this ticker. This can
                be stored and passed into `.remove(store_key=store_key)` later to
                easily stop this ticker later.

        Notes:
            The callback will be identified by type and stored either as
            as combination of serialized database object + methodname or
            as a python-path to the module + funcname. These strings will
            be combined iwth `interval` and `idstring` to define a
            unique storage key for saving. These must thus all be supplied
            when wanting to modify/remove the ticker later.

        """
        obj, path, callfunc = self._get_callback(callback)
        store_key = self._store_key(obj, path, interval, callfunc, idstring, persistent)
        kwargs["_obj"] = obj
        kwargs["_callback"] = callfunc  # either method-name or callable
        self.ticker_storage[store_key] = (args, kwargs)
        self.ticker_pool.add(store_key, *args, **kwargs)
        self.save()
        return store_key

    def remove(self, interval=60, callback=None, idstring="", persistent=True, store_key=None):
        """
        Remove ticker subscription from handler.

        Keyword Args:
            interval (int): Interval of ticker to remove.
            callback (callable function or method): Either a function or
                the method of a typeclassed object.
            idstring (str): Identifier id of ticker to remove.
            persistent (bool): Whether this ticker is persistent or not.
            store_key (str): If given, all other kwargs are ignored and only
                this is used to identify the ticker.

        Raises:
            KeyError: If no matching ticker was found to remove.

        Notes:
            The store-key is normally built from the interval/callback/idstring/persistent values;
            but if the `store_key` is explicitly given, this is used instead.

        """
        if isinstance(callback, int):
            raise RuntimeError(
                "TICKER_HANDLER.remove has changed: "
                "the interval is now the first argument, callback the second."
            )
        if not store_key:
            obj, path, callfunc = self._get_callback(callback)
            store_key = self._store_key(obj, path, interval, callfunc, idstring, persistent)
        to_remove = self.ticker_storage.pop(store_key, None)
        if to_remove:
            self.ticker_pool.remove(store_key)
            self.save()
        else:
            raise KeyError(f"No Ticker was found matching the store-key {store_key}.")

    def clear(self, interval=None):
        """
        Stop/remove tickers from handler.

        Args:
            interval (int, optional): Only stop tickers with this interval.

        Notes:
            This is the only supported way to kill tickers related to
            non-db objects.

        """
        self.ticker_pool.stop(interval)
        if interval:
            self.ticker_storage = dict(
                (store_key, store_value)
                for store_key, store_value in self.ticker_storage.items()
                if store_key[3] != interval
            )
        else:
            self.ticker_storage = {}
        self.save()

    def all(self, interval=None):
        """
        Get all ticker subscriptions.

        Args:
            interval (int, optional): Limit match to tickers with this interval.

        Returns:
            list or dict: If `interval` was given, this is a list of tickers using that interval.
            If `interval` was *not* given, this is a dict
            `{interval1: [ticker1, ticker2, ...],  ...}`

        """
        if interval is None:
            # return dict of all, ordered by interval
            return dict(
                (interval, ticker.subscriptions)
                for interval, ticker in self.ticker_pool.tickers.items()
            )
        else:
            # get individual interval
            ticker = self.ticker_pool.tickers.get(interval, None)
            if ticker:
                return {interval: ticker.subscriptions}
            return None

    def all_display(self):
        """
        Get all tickers on an easily displayable form.

        Returns:
            tickers (dict): A list of all storekeys

        """
        store_keys = []
        for ticker in self.ticker_pool.tickers.values():
            for (
                (objtup, callfunc, path, interval, idstring, persistent),
                (args, kwargs),
            ) in ticker.subscriptions.items():
                store_keys.append(
                    (kwargs.get("_obj", None), callfunc, path, interval, idstring, persistent)
                )
        return store_keys


# main tickerhandler
TICKER_HANDLER = TickerHandler()



"""
RSS parser for Evennia

This connects an RSS feed to an in-game Evennia channel, sending messages
to the channel whenever the feed updates.

"""
from django.conf import settings
from twisted.internet import task, threads

from evennia.server.session import Session
from evennia.utils import logger

RSS_ENABLED = settings.RSS_ENABLED
# RETAG = re.compile(r'<[^>]*?>')

if RSS_ENABLED:
    try:
        import feedparser
    except ImportError:
        raise ImportError(
            "RSS requires python-feedparser to be installed. Install or set RSS_ENABLED=False."
        )


class RSSReader(Session):
    """
    A simple RSS reader using the feedparser module.

    """

    def __init__(self, factory, url, rate):
        """
        Initialize the reader.

        Args:
            factory (RSSFactory): The protocol factory.
            url (str): The RSS url.
            rate (int): The seconds between RSS lookups.

        """
        self.url = url
        self.rate = rate
        self.factory = factory
        self.old_entries = {}

    def get_new(self):
        """
        Returns list of new items.

        """
        feed = feedparser.parse(self.url)
        new_entries = []
        for entry in feed["entries"]:
            idval = entry["id"] + entry.get("updated", "")
            if idval not in self.old_entries:
                self.old_entries[idval] = entry
                new_entries.append(entry)
        return new_entries

    def disconnect(self, reason=None):
        """
        Disconnect from feed.

        Args:
            reason (str, optional): Motivation for the disconnect.

        """
        if self.factory.task and self.factory.task.running:
            self.factory.task.stop()
        self.sessionhandler.disconnect(self)

    def _callback(self, new_entries, init):
        """
        Called when RSS returns.

        Args:
            new_entries (list): List of new RSS entries since last.
            init (bool): If this is a startup operation (at which
                point all entries are considered new).

        """
        if not init:
            # for initialization we just ignore old entries
            for entry in reversed(new_entries):
                self.data_in(entry)

    def data_in(self, text=None, **kwargs):
        """
        Data RSS -> Evennia.

        Keyword Args:
            text (str): Incoming text
            kwargs (any): Options from protocol.

        """
        self.sessionhandler.data_in(self, bot_data_in=text, **kwargs)

    def _errback(self, fail):
        "Report error"
        logger.log_err("RSS feed error: %s" % fail.value)

    def update(self, init=False):
        """
        Request the latest version of feed.

        Args:
            init (bool, optional): If this is an initialization call
                or not (during init, all entries are conidered new).

        Notes:
            This call is done in a separate thread to avoid blocking
            on slow connections.

        """
        return (
            threads.deferToThread(self.get_new)
            .addCallback(self._callback, init)
            .addErrback(self._errback)
        )


class RSSBotFactory(object):
    """
    Initializes new bots.
    """

    def __init__(self, sessionhandler, uid=None, url=None, rate=None):
        """
        Initialize the bot.

        Args:
            sessionhandler (PortalSessionHandler): The main sessionhandler object.
            uid (int): User id for the bot.
            url (str): The RSS URL.
            rate (int): How often for the RSS to request the latest RSS entries.

        """
        self.sessionhandler = sessionhandler
        self.url = url
        self.rate = rate
        self.uid = uid
        self.bot = RSSReader(self, url, rate)
        self.task = None

    def start(self):
        """
        Called by portalsessionhandler. Starts the bot.

        """

        def errback(fail):
            logger.log_err(fail.value)

        # set up session and connect it to sessionhandler
        self.bot.init_session("rssbot", self.url, self.sessionhandler)
        self.bot.uid = self.uid
        self.bot.logged_in = True
        self.sessionhandler.connect(self.bot)

        # start repeater task
        self.bot.update(init=True)
        self.task = task.LoopingCall(self.bot.update)
        if self.rate:
            self.task.start(self.rate, now=False).addErrback(errback)



"""

Helpers and resources for the map system.

"""

BIGVAL = 999999999999

REVERSE_DIRECTIONS = {
    "n": "s",
    "ne": "sw",
    "e": "w",
    "se": "nw",
    "s": "n",
    "sw": "ne",
    "w": "e",
    "nw": "se",
}

MAPSCAN = {
    "n": (0, 1),
    "ne": (1, 1),
    "e": (1, 0),
    "se": (1, -1),
    "s": (0, -1),
    "sw": (-1, -1),
    "w": (-1, 0),
    "nw": (-1, 1),
}

# errors for Map system


class MapError(RuntimeError):
    def __init__(self, error="", node_or_link=None):
        prefix = ""
        if node_or_link:
            prefix = (
                f"{node_or_link.__class__.__name__} '{node_or_link.symbol}' "
                f"at XYZ=({node_or_link.X:g},{node_or_link.Y:g},{node_or_link.Z}) "
            )
        self.node_or_link = node_or_link
        self.message = f"{prefix}{error}"
        super().__init__(self.message)


class MapParserError(MapError):
    pass


class MapTransition(RuntimeWarning):
    """
    Used when signaling to the parser that a link
    leads to another map.

    """

    pass



"""
Admin views.

"""

from django.contrib.admin.sites import site
from django.contrib.admin.views.decorators import staff_member_required
from django.shortcuts import render

from evennia.accounts.models import AccountDB


@staff_member_required
def evennia_admin(request):
    """
    Helpful Evennia-specific admin page.

    """
    return render(request, "admin/frontpage.html", {"accountdb": AccountDB})


def admin_wrapper(request):
    """
    Wrapper that allows us to properly use the base Django admin site, if needed.

    """
    return staff_member_required(site.index)(request)



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("objects", "0004_auto_20150118_1622"), ("comms", "0005_auto_20150223_1517")]

    operations = [
        migrations.AddField(
            model_name="channeldb",
            name="db_object_subscriptions",
            field=models.ManyToManyField(
                related_name="object_subscription_set",
                null=True,
                verbose_name="subscriptions",
                to="objects.ObjectDB",
                db_index=True,
            ),
            preserve_default=True,
        )
    ]



# -*- coding: utf-8 -*-

"""
Evennia World Builder

Contribution - Cloud_Keeper 2016

Build a map from a 2D ASCII map.

This is a command which takes two inputs:

≈≈≈≈≈
≈♣n♣≈   MAP_LEGEND = {("♣", "♠"): build_forest,
≈∩▲∩≈                 ("∩", "n"): build_mountains,
≈♠n♠≈                 ("▲"): build_temple}
≈≈≈≈≈

A string of ASCII characters representing a map and a dictionary of functions
containing build instructions. The characters of the map are iterated over and
compared to a list of trigger characters. When a match is found the
corresponding function is executed generating the rooms, exits and objects as
defined by the users build instructions. If a character is not a match to
a provided trigger character (including spaces) it is simply skipped and the
process continues.

For instance, the above map represents a temple (▲) amongst mountains (n,∩)
in a forest (♣,♠) on an island surrounded by water (≈). Each character on the
first line is iterated over but as there is no match with our MAP_LEGEND it
is skipped. On the second line it finds "♣" which is a match and so the
`build_forest` function is called. Next the `build_mountains` function is
called and so on until the map is completed. Building instructions are passed
the following arguments:
    x         - The rooms position on the maps x axis
    y         - The rooms position on the maps y axis
    caller    - The account calling the command
    iteration - The current iterations number (0, 1 or 2)
    room_dict - A dictionary containing room references returned by build
                functions where tuple coordinates are the keys (x, y).
                ie room_dict[(2, 2)] will return the temple room above.

Building functions should return the room they create. By default these rooms
are used to create exits between valid adjacent rooms to the north, south,
east and west directions. This behaviour can turned off with the use of switch
arguments. In addition to turning off automatic exit generation the switches
allow the map to be iterated over a number of times. This is important for
something like custom exit building. Exits require a reference to both the
exits location and the exits destination. During the first iteration it is
possible that an exit is created pointing towards a destination that
has not yet been created resulting in error. By iterating over the map twice
the rooms can be created on the first iteration and room reliant code can be
be used on the second iteration. The iteration number and a dictionary of
references to rooms previously created is passed to the build commands.

You then call the command in-game using the path to the MAP and MAP_LEGEND vars
The path you provide is relative to the evennia or mygame folder.

## Installation:

Use by importing and including the command in your default_cmdsets module.
For example:

```python
    # mygame/commands/default_cmdsets.py

    from evennia.contrib.grid import mapbuilder

    ...

    self.add(mapbuilder.CmdMapBuilder())
```


## Usage

    mapbuilder[/switch] <path.to.file.MAPNAME> <path.to.file.MAP_LEGEND>

    one - execute build instructions once without automatic exit creation.
    two - execute build instructions twice without automatic exit creation.

## Example

    mapbuilder world.gamemap.MAP world.maplegend.MAP_LEGEND
    mapbuilder evennia.contrib.grid.mapbuilder.EXAMPLE1_MAP EXAMPLE1_LEGEND
    mapbuilder/two evennia.contrib.grid.mapbuilder.EXAMPLE2_MAP EXAMPLE2_LEGEND
            (Legend path defaults to map path)

Below are two examples showcasing the use of automatic exit generation and
custom exit generation. Whilst located, and can be used, from this module for
convenience The below example code should be in mymap.py in mygame/world.

# Example One:

```python
# @mapbuilder evennia.contrib.grid.mapbuilder.EXAMPLE1_MAP EXAMPLE1_LEGEND

# -*- coding: utf-8 -*-

# Add the necessary imports for your instructions here.
from evennia import create_object
from typeclasses import rooms, exits
from random import randint
import random


# A map with a temple (▲) amongst mountains (n,∩) in a forest (♣,♠) on an
# island surrounded by water (≈). By giving no instructions for the water
# characters we effectively skip it and create no rooms for those squares.
EXAMPLE1_MAP = '''
≈≈≈≈≈
≈♣n♣≈
≈∩▲∩≈
≈♠n♠≈
≈≈≈≈≈
'''

def example1_build_forest(x, y, **kwargs):
    '''A basic example of build instructions. Make sure to include **kwargs
    in the arguments and return an instance of the room for exit generation.'''

    # Create a room and provide a basic description.
    room = create_object(rooms.Room, key="forest" + str(x) + str(y))
    room.db.desc = "Basic forest room."

    # Send a message to the account
    kwargs["caller"].msg(room.key + " " + room.dbref)

    # This is generally mandatory.
    return room


def example1_build_mountains(x, y, **kwargs):
    '''A room that is a little more advanced'''

    # Create the room.
    room = create_object(rooms.Room, key="mountains" + str(x) + str(y))

    # Generate a description by randomly selecting an entry from a list.
    room_desc = [
        "Mountains as far as the eye can see",
        "Your path is surrounded by sheer cliffs",
        "Haven't you seen that rock before?",
    ]
    room.db.desc = random.choice(room_desc)

    # Create a random number of objects to populate the room.
    for i in range(randint(0, 3)):
        rock = create_object(key="Rock", location=room)
        rock.db.desc = "An ordinary rock."

    # Send a message to the account
    kwargs["caller"].msg(room.key + " " + room.dbref)

    # This is generally mandatory.
    return room


def example1_build_temple(x, y, **kwargs):
    '''A unique room that does not need to be as general'''

    # Create the room.
    room = create_object(rooms.Room, key="temple" + str(x) + str(y))

    # Set the description.
    room.db.desc = (
        "In what, from the outside, appeared to be a grand and "
        "ancient temple you've somehow found yourself in the the "
        "Evennia Inn! It consists of one large room filled with "
        "tables. The bardisk extends along the east wall, where "
        "multiple barrels and bottles line the shelves. The "
        "barkeep seems busy handing out ale and chatting with "
        "the patrons, which are a rowdy and cheerful lot, "
        "keeping the sound level only just below thunderous. "
        "This is a rare spot of mirth on this dread moor."
    )

    # Send a message to the account
    kwargs["caller"].msg(room.key + " " + room.dbref)

    # This is generally mandatory.
    return room


# Include your trigger characters and build functions in a legend dict.
EXAMPLE1_LEGEND = {
    ("♣", "♠"): example1_build_forest,
    ("∩", "n"): example1_build_mountains,
    ("▲"): example1_build_temple,
}
```

# Example two

```python
# @mapbuilder/two evennia.contrib.grid.mapbuilder.EXAMPLE2_MAP EXAMPLE2_LEGEND

# -*- coding: utf-8 -*-

# Add the necessary imports for your instructions here.
# from evennia import create_object
# from typeclasses import rooms, exits
# from evennia.utils import utils
# from random import randint
# import random

# This is the same layout as Example 1 but included are characters for exits.
# We can use these characters to determine which rooms should be connected.
EXAMPLE2_MAP = '''
≈ ≈ ≈ ≈ ≈

≈ ♣-♣-♣ ≈
  |   |
≈ ♣ ♣ ♣ ≈
  | | |
≈ ♣-♣-♣ ≈

≈ ≈ ≈ ≈ ≈
'''

def example2_build_forest(x, y, **kwargs):
    '''A basic room'''
    # If on anything other than the first iteration - Do nothing.
    if kwargs["iteration"] > 0:
        return None

    room = create_object(rooms.Room, key="forest" + str(x) + str(y))
    room.db.desc = "Basic forest room."

    kwargs["caller"].msg(room.key + " " + room.dbref)

    return room


def example2_build_verticle_exit(x, y, **kwargs):
    '''Creates two exits to and from the two rooms north and south.'''
    # If on the first iteration - Do nothing.
    if kwargs["iteration"] == 0:
        return

    north_room = kwargs["room_dict"][(x, y - 1)]
    south_room = kwargs["room_dict"][(x, y + 1)]

    # create exits in the rooms
    create_object(
        exits.Exit, key="south", aliases=["s"], location=north_room, destination=south_room
    )

    create_object(
        exits.Exit, key="north", aliases=["n"], location=south_room, destination=north_room
    )

    kwargs["caller"].msg("Connected: " + north_room.key + " & " + south_room.key)


def example2_build_horizontal_exit(x, y, **kwargs):
    '''Creates two exits to and from the two rooms east and west.'''
    # If on the first iteration - Do nothing.
    if kwargs["iteration"] == 0:
        return

    west_room = kwargs["room_dict"][(x - 1, y)]
    east_room = kwargs["room_dict"][(x + 1, y)]

    create_object(exits.Exit, key="east", aliases=["e"], location=west_room, destination=east_room)

    create_object(exits.Exit, key="west", aliases=["w"], location=east_room, destination=west_room)

    kwargs["caller"].msg("Connected: " + west_room.key + " & " + east_room.key)


# Include your trigger characters and build functions in a legend dict.
EXAMPLE2_LEGEND = {
    ("♣", "♠"): example2_build_forest,
    ("|"): example2_build_verticle_exit,
    ("-"): example2_build_horizontal_exit,
}

```

---
"""
from django.conf import settings
from typeclasses import exits

from evennia import create_object
from evennia.utils import utils

COMMAND_DEFAULT_CLASS = utils.class_from_module(settings.COMMAND_DEFAULT_CLASS)


# Helper function for readability.
def _map_to_list(game_map):
    """
    Splits multi line map string into list of rows.

    Args:
        game_map (str): An ASCII map

    Returns:
        list (list): The map split into rows

    """
    return game_map.split("\n")


def build_map(caller, game_map, legend, iterations=1, build_exits=True):
    """
    Receives the fetched map and legend vars provided by the player.

    Args:
        caller (Object): The creator of the map.
        game_map (str): An ASCII map string.
        legend (dict): Mapping of map symbols to object types.
        iterations (int): The number of iteration passes.
        build_exits (bool): Create exits between new rooms.

    Notes:
        The map
        is iterated over character by character, comparing it to the trigger
        characters in the legend var and executing the build instructions on
        finding a match. The map is iterated over according to the `iterations`
        value and exits are optionally generated between adjacent rooms according
        to the `build_exits` value.

    """

    # Split map string to list of rows and create reference list.
    caller.msg("Creating Map...")
    caller.msg(game_map)
    game_map = _map_to_list(game_map)

    # Create a reference dictionary which be passed to build functions and
    # will store obj returned by build functions so objs can be referenced.
    room_dict = {}

    caller.msg("Creating Landmass...")
    for iteration in range(iterations):
        for y in range(len(game_map)):
            for x in range(len(game_map[y])):
                for key in legend:
                    # obs - we must use == for strings
                    if game_map[y][x] == key:
                        room = legend[key](
                            x, y, iteration=iteration, room_dict=room_dict, caller=caller
                        )
                        if iteration == 0:
                            room_dict[(x, y)] = room

    if build_exits:
        # Creating exits. Assumes single room object in dict entry
        caller.msg("Connecting Areas...")
        for loc_key, location in room_dict.items():
            x = loc_key[0]
            y = loc_key[1]

            # north
            if (x, y - 1) in room_dict:
                if room_dict[(x, y - 1)]:
                    create_object(
                        exits.Exit,
                        key="north",
                        aliases=["n"],
                        location=location,
                        destination=room_dict[(x, y - 1)],
                    )

            # east
            if (x + 1, y) in room_dict:
                if room_dict[(x + 1, y)]:
                    create_object(
                        exits.Exit,
                        key="east",
                        aliases=["e"],
                        location=location,
                        destination=room_dict[(x + 1, y)],
                    )

            # south
            if (x, y + 1) in room_dict:
                if room_dict[(x, y + 1)]:
                    create_object(
                        exits.Exit,
                        key="south",
                        aliases=["s"],
                        location=location,
                        destination=room_dict[(x, y + 1)],
                    )

            # west
            if (x - 1, y) in room_dict:
                if room_dict[(x - 1, y)]:
                    create_object(
                        exits.Exit,
                        key="west",
                        aliases=["w"],
                        location=location,
                        destination=room_dict[(x - 1, y)],
                    )

    caller.msg("Map Created.")


# access command


class CmdMapBuilder(COMMAND_DEFAULT_CLASS):
    """
    Build a map from a 2D ASCII map.

    Usage:
        @mapbuilder[/switch] <path.to.file.MAPNAME> <path.to.file.MAP_LEGEND>

    Switches:
      one - execute build instructions once without automatic exit creation
      two - execute build instructions twice without automatic exit creation

    Example:
        @mapbuilder world.gamemap.MAP world.maplegend.MAP_LEGEND
        @mapbuilder evennia.contrib.grid.mapbuilder.EXAMPLE1_MAP EXAMPLE1_LEGEND
        @mapbuilder/two evennia.contrib.grid.mapbuilder.EXAMPLE2_MAP EXAMPLE2_LEGEND
                (Legend path defaults to map path)

    This is a command which takes two inputs:
    A string of ASCII characters representing a map and a dictionary of
    functions containing build instructions. The characters of the map are
    iterated over and compared to a list of trigger characters. When a match
    is found the corresponding function is executed generating the rooms,
    exits and objects as defined by the users build instructions. If a
    character is not a match to a provided trigger character (including spaces)
    it is simply skipped and the process continues. By default exits are
    automatically generated but is turned off by switches which also determines
    how many times the map is iterated over.
    """

    key = "@mapbuilder"
    aliases = ["@buildmap"]
    locks = "cmd:superuser()"
    help_category = "Building"

    def func(self):
        """Starts the processor."""

        caller = self.caller
        args = self.args.split()

        # Check if arguments passed.
        if not self.args or (len(args) != 2):
            caller.msg("Usage: @mapbuilder <path.to.module.VARNAME> " "<path.to.module.MAP_LEGEND>")
            return

        # Set up base variables.
        game_map = None
        legend = None

        # OBTAIN MAP FROM MODULE

        # Breaks down path_to_map into [PATH, VARIABLE]
        path_to_map = args[0]
        path_to_map = path_to_map.rsplit(".", 1)

        try:
            # Retrieves map variable from module or raises error.
            game_map = utils.variable_from_module(path_to_map[0], path_to_map[1])
            if not game_map:
                raise ValueError(
                    "Command Aborted!\n"
                    "Path to map variable failed.\n"
                    "Usage: @mapbuilder <path.to.module."
                    "VARNAME> <path.to.module.MAP_LEGEND>"
                )

        except Exception as exc:
            # Or relays error message if fails.
            caller.msg(exc)
            return

        # OBTAIN MAP_LEGEND FROM MODULE

        # Breaks down path_to_legend into [PATH, VARIABLE]
        path_to_legend = args[1]
        path_to_legend = path_to_legend.rsplit(".", 1)

        # If no path given default to path_to_map's path
        if len(path_to_legend) == 1:
            path_to_legend.insert(0, path_to_map[0])

        try:
            # Retrieves legend variable from module or raises error if fails.
            legend = utils.variable_from_module(path_to_legend[0], path_to_legend[1])
            if not legend:
                raise ValueError(
                    "Command Aborted!\n"
                    "Path to legend variable failed.\n"
                    "Usage: @mapbuilder <path.to.module."
                    "VARNAME> <path.to.module.MAP_LEGEND>"
                )

        except Exception as exc:
            # Or relays error message if fails.
            caller.msg(exc)
            return

        # Set up build_map arguments from switches
        iterations = 1
        build_exits = True

        if "one" in self.switches:
            build_exits = False

        if "two" in self.switches:
            iterations = 2
            build_exits = False

        # Pass map and legend to the build function.
        build_map(caller, game_map, legend, iterations, build_exits)



"""
Basic class for NPC that makes use of an LLM (Large Language Model) to generate replies.

It comes with a `talk` command; use `talk npc <something>` to talk to the NPC. The NPC will
respond using the LLM response.

Makes use of the LLMClient for communicating with the server. The NPC will also
echo a 'thinking...' message if the LLM server takes too long to respond.


"""

from collections import defaultdict
from random import choice

from django.conf import settings
from twisted.internet import reactor, task
from twisted.internet.defer import CancelledError, inlineCallbacks

from evennia import AttributeProperty, Command, DefaultCharacter
from evennia.utils.utils import make_iter

from .llm_client import LLMClient

# fallback if not specified anywhere else. Check order is
# npc.db.prompt_prefix, npcClass.prompt_prefix, then settings.LLM_PROMPT_PREFIX, then this
DEFAULT_PROMPT_PREFIX = (
    "You are roleplaying as {name}, a {desc} existing in {location}. "
    "Answer with short sentences. Only respond as {name} would. "
    "From here on, the conversation between {name} and {character} begins."
)


class LLMNPC(DefaultCharacter):
    """An NPC that uses the LLM server to generate its responses. If the server is slow, it will
    echo a thinking message to the character while it waits for a response."""

    # use this to override the prefix per class. Assign an Attribute to override per-instance.
    prompt_prefix = None

    response_template = AttributeProperty(
        "$You() $conj(say) (to $You(character)): {response}", autocreate=False
    )
    thinking_timeout = AttributeProperty(2, autocreate=False)  # seconds
    thinking_messages = AttributeProperty(
        [
            "{name} thinks about what you said ...",
            "{name} ponders your words ...",
            "{name} ponders ...",
        ],
        autocreate=False,
    )

    max_chat_memory_size = AttributeProperty(25, autocreate=False)
    # this is a store of {character: [chat, chat, ...]}
    chat_memory = AttributeProperty(defaultdict(list))

    @property
    def llm_client(self):
        if not self.ndb.llm_client:
            self.ndb.llm_client = LLMClient()
        return self.ndb.llm_client

    @property
    def llm_prompt_prefix(self):
        """get prefix, first from Attribute, then from class variable,
        then from settings, then from default"""
        return self.attributes.get(
            "prompt_prefix",
            default=getattr(
                settings, "LLM_PROMPT_PREFIX", self.prompt_prefix or DEFAULT_PROMPT_PREFIX
            ),
        )

    def _add_to_memory(self, character, who_talked, speech):
        """Add a person's speech to the memory. This is stored as name: chat for the LLM."""
        memory = self.chat_memory[character]
        memory.append(f"{who_talked.get_display_name(self)}: {speech}")

        # trim the memory if it's getting too long in order to save space
        memory = memory[-self.max_chat_memory_size :]
        self.chat_memory[character] = memory

    def build_prompt(self, character, speech):
        """
        Build the prompt to send to the LLM server.

        Args:
            character (Object): The one talking to the NPC.
            speech (str): The latest speech from the character.

        Returns:
            str: The prompt to return.

        """
        name = self.get_display_name(character)
        charname = character.get_display_name(self)
        memory = self.chat_memory[character]

        # get starting prompt
        prompt = self.llm_prompt_prefix.format(
            name=name,
            desc=self.db.desc or "someone",
            location=self.location.key if self.location else "the void",
            character=charname,
        )
        prompt += "\n" + "\n".join(mem for mem in memory)
        return prompt

    @inlineCallbacks
    def at_talked_to(self, speech, character):
        """Called when this NPC is talked to by a character."""

        def _respond(response):
            """Async handling of the server response"""

            if thinking_defer and not thinking_defer.called:
                # abort the thinking message if we were fast enough
                thinking_defer.cancel()

            if response:
                # remember this response
                self._add_to_memory(character, self, response)
            else:
                response = "... I'm sorry, I was distracted. Can you repeat?"

            response = self.response_template.format(
                name=self.get_display_name(character), response=response
            )

            # tell the character about it
            if character.location:
                character.location.msg_contents(
                    response,
                    mapping={"character": character},
                    from_obj=self,
                )
            else:
                # fallback if character is not in a location
                character.msg(f"{self.get_display_name(character)} says, {response}")

        # if response takes too long, note that the NPC is thinking.

        def _echo_thinking_message():
            """Echo a random thinking message to the character"""
            thinking_message = choice(
                make_iter(self.db.thinking_messages or self.thinking_messages)
            )
            if character.location:
                thinking_message = thinking_message.format(name="$You()")
                character.location.msg_contents(thinking_message, from_obj=self)
            else:
                thinking_message = thinking_message.format(name=self.get_display_name(character))
                character.msg(thinking_message)

        def _handle_cancel_error(failure):
            """Suppress task-cancel errors only"""
            failure.trap(CancelledError)

        thinking_defer = task.deferLater(
            reactor, self.thinking_timeout, _echo_thinking_message
        ).addErrback(_handle_cancel_error)

        # remember latest input in memory, so it's included in the prompt
        self._add_to_memory(character, character, speech)

        # build the prompt
        prompt = self.build_prompt(character, speech)

        # get the response from the LLM server
        yield self.llm_client.get_response(prompt).addCallback(_respond)


class CmdLLMTalk(Command):
    """
    Talk to an NPC

    Usage:
        talk npc <something>
        talk npc with spaces in name = <something>

    """

    key = "talk"

    def parse(self):
        args = self.args.strip()
        if "s=" in args:
            name, *speech = args.split("=", 1)
        else:
            name, *speech = args.split(" ", 1)
        self.target_name = name
        self.speech = speech[0] if speech else ""

    def func(self):
        if not self.target_name:
            self.caller.msg("Talk to who?")
            return

        location = self.caller.location
        target = self.caller.search(self.target_name)
        if not target:
            return
        if location:
            location.msg_contents(
                f"$You() $conj(say) (to $You(target)): {self.speech}",
                mapping={"target": target},
                from_obj=self.caller,
            )
        if hasattr(target, "at_talked_to"):
            target.at_talked_to(self.speech, self.caller)
        else:
            self.caller.msg(f"{target.key} doesn't seem to want to talk to you.")



"""
Set a more useful description on the Api root.

"""

from rest_framework import routers


class EvenniaAPIRoot(routers.APIRootView):
    """
    Root of the Evennia API tree.

    """

    pass


class APIRootRouter(routers.DefaultRouter):
    APIRootView = EvenniaAPIRoot



"""
Simple turn-based combat system with spell casting

Contrib - Tim Ashley Jenkins 2017, Refactor by Griatch, 2022

This is a version of the 'turnbattle' contrib that includes a basic,
expandable framework for a 'magic system', whereby players can spend
a limited resource (MP) to achieve a wide variety of effects, both in
and out of combat. This does not have to strictly be a system for
magic - it can easily be re-flavored to any other sort of resource
based mechanic, like psionic powers, special moves and stamina, and
so forth.

In this system, spells are learned by name with the 'learnspell'
command, and then used with the 'cast' command. Spells can be cast in or
out of combat - some spells can only be cast in combat, some can only be
cast outside of combat, and some can be cast any time. However, if you
are in combat, you can only cast a spell on your turn, and doing so will
typically use an action (as specified in the spell's funciton).

Spells are defined at the end of the module in a database that's a
dictionary of dictionaries - each spell is matched by name to a function,
along with various parameters that restrict when the spell can be used and
what the spell can be cast on. Included is a small variety of spells that
damage opponents and heal HP, as well as one that creates an object.

Because a spell can call any function, a spell can be made to do just
about anything at all. The SPELLS dictionary at the bottom of the module
even allows kwargs to be passed to the spell function, so that the same
function can be re-used for multiple similar spells.

Spells in this system work on a very basic resource: MP, which is spent
when casting spells and restored by resting. It shouldn't be too difficult
to modify this system to use spell slots, some physical fuel or resource,
or whatever else your game requires.

To install and test, import this module's TBMagicCharacter object into
your game's character.py module:

    from evennia.contrib.game_systems.turnbattle.tb_magic import TBMagicCharacter

And change your game's character typeclass to inherit from TBMagicCharacter
instead of the default:

    class Character(TBMagicCharacter):

Note: If your character already existed you need to also make sure
to re-run the creation hooks on it to set the needed Attributes.
Use `update self` to try on yourself or use py to call `at_object_creation()`
on all existing Characters.


Next, import this module into your default_cmdsets.py module:

    from evennia.contrib.game_systems.turnbattle import tb_magic

And add the battle command set to your default command set:

    #
    # any commands you add below will overload the default ones.
    #
    self.add(tb_magic.BattleCmdSet())

This module is meant to be heavily expanded on, so you may want to copy it
to your game's 'world' folder and modify it there rather than importing it
in your game and using it as-is.
"""

from random import randint

from evennia import Command, DefaultScript, create_object, default_cmds
from evennia.commands.default.muxcommand import MuxCommand
from evennia.utils.logger import log_trace

from . import tb_basic

"""
----------------------------------------------------------------------------
SPELL FUNCTIONS START HERE
----------------------------------------------------------------------------

These are the functions that are called by the 'cast' command to perform the
effects of various spells. Which spells execute which functions and what
parameters are passed to them are specified at the bottom of the module, in
the 'SPELLS' dictionary.

All of these functions take the same arguments:
    caster (obj): Character casting the spell
    spell_name (str): Name of the spell being cast
    targets (list): List of objects targeted by the spell
    cost (int): MP cost of casting the spell

These functions also all accept **kwargs, and how these are used is specified
in the docstring for each function.
"""


class MagicCombatRules(tb_basic.BasicCombatRules):
    def spell_healing(self, caster, spell_name, targets, cost, **kwargs):
        """
        Spell that restores HP to a target or targets.

        kwargs:
            healing_range (tuple): Minimum and maximum amount healed to
                each target. (20, 40) by default.
        """
        spell_msg = "%s casts %s!" % (caster, spell_name)

        min_healing = 20
        max_healing = 40

        # Retrieve healing range from kwargs, if present
        if "healing_range" in kwargs:
            min_healing = kwargs["healing_range"][0]
            max_healing = kwargs["healing_range"][1]

        for character in targets:
            to_heal = randint(min_healing, max_healing)  # Restore 20 to 40 hp
            if character.db.hp + to_heal > character.db.max_hp:
                to_heal = character.db.max_hp - character.db.hp  # Cap healing to max HP
            character.db.hp += to_heal
            spell_msg += " %s regains %i HP!" % (character, to_heal)

        caster.db.mp -= cost  # Deduct MP cost

        caster.location.msg_contents(spell_msg)  # Message the room with spell results

        if self.is_in_combat(caster):  # Spend action if in combat
            self.spend_action(caster, 1, action_name="cast")

    def spell_attack(self, caster, spell_name, targets, cost, **kwargs):
        """
        Spell that deals damage in combat. Similar to resolve_attack.

        kwargs:
            attack_name (tuple): Single and plural describing the sort of
                attack or projectile that strikes each enemy.
            damage_range (tuple): Minimum and maximum damage dealt by the
                spell. (10, 20) by default.
            accuracy (int): Modifier to the spell's attack roll, determining
                an increased or decreased chance to hit. 0 by default.
            attack_count (int): How many individual attacks are made as part
                of the spell. If the number of attacks exceeds the number of
                targets, the first target specified will be attacked more
                than once. Just 1 by default - if the attack_count is less
                than the number targets given, each target will only be
                attacked once.
        """
        spell_msg = "%s casts %s!" % (caster, spell_name)

        atkname_single = "The spell"
        atkname_plural = "spells"
        min_damage = 10
        max_damage = 20
        accuracy = 0
        attack_count = 1

        # Retrieve some variables from kwargs, if present
        if "attack_name" in kwargs:
            atkname_single = kwargs["attack_name"][0]
            atkname_plural = kwargs["attack_name"][1]
        if "damage_range" in kwargs:
            min_damage = kwargs["damage_range"][0]
            max_damage = kwargs["damage_range"][1]
        if "accuracy" in kwargs:
            accuracy = kwargs["accuracy"]
        if "attack_count" in kwargs:
            attack_count = kwargs["attack_count"]

        to_attack = []
        # If there are more attacks than targets given, attack first target multiple times
        if len(targets) < attack_count:
            to_attack = to_attack + targets
            extra_attacks = attack_count - len(targets)
            for n in range(extra_attacks):
                to_attack.insert(0, targets[0])
        else:
            to_attack = to_attack + targets

        # Set up dictionaries to track number of hits and total damage
        total_hits = {}
        total_damage = {}
        for fighter in targets:
            total_hits.update({fighter: 0})
            total_damage.update({fighter: 0})

        # Resolve attack for each target
        for fighter in to_attack:
            attack_value = randint(1, 100) + accuracy  # Spell attack roll
            defense_value = self.get_defense(caster, fighter)
            if attack_value >= defense_value:
                spell_dmg = randint(min_damage, max_damage)  # Get spell damage
                total_hits[fighter] += 1
                total_damage[fighter] += spell_dmg

        for fighter in targets:
            # Construct combat message
            if total_hits[fighter] == 0:
                spell_msg += " The spell misses %s!" % fighter
            elif total_hits[fighter] > 0:
                attack_count_str = atkname_single + " hits"
                if total_hits[fighter] > 1:
                    attack_count_str = "%i %s hit" % (total_hits[fighter], atkname_plural)
                spell_msg += " %s %s for %i damage!" % (
                    attack_count_str,
                    fighter,
                    total_damage[fighter],
                )

        caster.db.mp -= cost  # Deduct MP cost

        caster.location.msg_contents(spell_msg)  # Message the room with spell results

        for fighter in targets:
            # Apply damage
            self.apply_damage(fighter, total_damage[fighter])
            # If fighter HP is reduced to 0 or less, call at_defeat.
            if fighter.db.hp <= 0:
                self.at_defeat(fighter)

        if self.is_in_combat(caster):  # Spend action if in combat
            self.spend_action(caster, 1, action_name="cast")

    def spell_conjure(self, caster, spell_name, targets, cost, **kwargs):
        """
        Spell that creates an object.

        kwargs:
            obj_key (str): Key of the created object.
            obj_desc (str): Desc of the created object.
            obj_typeclass (str): Typeclass path of the object.

        If you want to make more use of this particular spell funciton,
        you may want to modify it to use the spawner (in evennia.utils.spawner)
        instead of creating objects directly.
        """

        obj_key = "a nondescript object"
        obj_desc = "A perfectly generic object."
        obj_typeclass = "evennia.objects.objects.DefaultObject"

        # Retrieve some variables from kwargs, if present
        if "obj_key" in kwargs:
            obj_key = kwargs["obj_key"]
        if "obj_desc" in kwargs:
            obj_desc = kwargs["obj_desc"]
        if "obj_typeclass" in kwargs:
            obj_typeclass = kwargs["obj_typeclass"]

        conjured_obj = create_object(
            obj_typeclass, key=obj_key, location=caster.location
        )  # Create object
        conjured_obj.db.desc = obj_desc  # Add object desc

        caster.db.mp -= cost  # Deduct MP cost

        # Message the room to announce the creation of the object
        caster.location.msg_contents(
            "%s casts %s, and %s appears!" % (caster, spell_name, conjured_obj)
        )


COMBAT_RULES = MagicCombatRules()


"""
----------------------------------------------------------------------------
SPELL DEFINITIONS START HERE
----------------------------------------------------------------------------
In this section, each spell is matched to a function, and given parameters
that determine its MP cost, valid type and number of targets, and what
function casting the spell executes.

This data is given as a dictionary of dictionaries - the key of each entry
is the spell's name, and the value is a dictionary of various options and
parameters, some of which are required and others which are optional.

Required values for spells:

    cost (int): MP cost of casting the spell
    target (str): Valid targets for the spell. Can be any of:
        "none" - No target needed
        "self" - Self only
        "any" - Any object
        "anyobj" - Any object that isn't a character
        "anychar" - Any character
        "other" - Any object excluding the caster
        "otherchar" - Any character excluding the caster
    spellfunc (callable): Function that performs the action of the spell.
        Must take the following arguments: caster (obj), spell_name (str),
        targets (list), and cost (int), as well as **kwargs.

Optional values for spells:

    combat_spell (bool): If the spell can be cast in combat. True by default.
    noncombat_spell (bool): If the spell can be cast out of combat. True by default.
    max_targets (int): Maximum number of objects that can be targeted by the spell.
        1 by default - unused if target is "none" or "self"

Any other values specified besides the above will be passed as kwargs to 'spellfunc'.
You can use kwargs to effectively re-use the same function for different but similar
spells - for example, 'magic missile' and 'flame shot' use the same function, but
behave differently, as they have different damage ranges, accuracy, amount of attacks
made as part of the spell, and so forth. If you make your spell functions flexible
enough, you can make a wide variety of spells just by adding more entries to this
dictionary.
"""

SPELLS = {
    "magic missile": {
        "spellfunc": COMBAT_RULES.spell_attack,
        "target": "otherchar",
        "cost": 3,
        "noncombat_spell": False,
        "max_targets": 3,
        "attack_name": ("A bolt", "bolts"),
        "damage_range": (4, 7),
        "accuracy": 999,
        "attack_count": 3,
    },
    "flame shot": {
        "spellfunc": COMBAT_RULES.spell_attack,
        "target": "otherchar",
        "cost": 3,
        "noncombat_spell": False,
        "attack_name": ("A jet of flame", "jets of flame"),
        "damage_range": (25, 35),
    },
    "cure wounds": {"spellfunc": COMBAT_RULES.spell_healing, "target": "anychar", "cost": 5},
    "mass cure wounds": {
        "spellfunc": COMBAT_RULES.spell_healing,
        "target": "anychar",
        "cost": 10,
        "max_targets": 5,
    },
    "full heal": {
        "spellfunc": COMBAT_RULES.spell_healing,
        "target": "anychar",
        "cost": 12,
        "healing_range": (100, 100),
    },
    "cactus conjuration": {
        "spellfunc": COMBAT_RULES.spell_conjure,
        "target": "none",
        "cost": 2,
        "combat_spell": False,
        "obj_key": "a cactus",
        "obj_desc": "An ordinary green cactus with little spines.",
    },
}


"""
----------------------------------------------------------------------------
OPTIONS
----------------------------------------------------------------------------
"""

TURN_TIMEOUT = 30  # Time before turns automatically end, in seconds
ACTIONS_PER_TURN = 1  # Number of actions allowed per turn


"""
----------------------------------------------------------------------------
CHARACTER TYPECLASS
----------------------------------------------------------------------------
"""


class TBMagicCharacter(tb_basic.TBBasicCharacter):
    """
    A character able to participate in turn-based combat. Has attributes for current
    and maximum HP, access to combat commands and magic.

    """

    rules = COMBAT_RULES

    def at_object_creation(self):
        """
        Called once, when this object is first created. This is the
        normal hook to overload for most object types.

        Adds attributes for a character's current and maximum HP.
        We're just going to set this value at '100' by default.

        You may want to expand this to include various 'stats' that
        can be changed at creation and factor into combat calculations.
        """
        self.db.max_hp = 100  # Set maximum HP to 100
        self.db.hp = self.db.max_hp  # Set current HP to maximum
        self.db.spells_known = []  # Set empty spells known list
        self.db.max_mp = 20  # Set maximum MP to 20
        self.db.mp = self.db.max_mp  # Set current MP to maximum


"""
----------------------------------------------------------------------------
SCRIPTS START HERE
----------------------------------------------------------------------------
"""


class TBMagicTurnHandler(tb_basic.TBBasicTurnHandler):
    """
    This is the script that handles the progression of combat through turns.
    On creation (when a fight is started) it adds all combat-ready characters
    to its roster and then sorts them into a turn order. There can only be one
    fight going on in a single room at a time, so the script is assigned to a
    room as its object.

    Fights persist until only one participant is left with any HP or all
    remaining participants choose to end the combat with the 'disengage' command.
    """

    rules = COMBAT_RULES


"""
----------------------------------------------------------------------------
COMMANDS START HERE
----------------------------------------------------------------------------
"""


class CmdFight(tb_basic.CmdFight):
    """
    Starts a fight with everyone in the same room as you.

    Usage:
      fight

    When you start a fight, everyone in the room who is able to
    fight is added to combat, and a turn order is randomly rolled.
    When it's your turn, you can attack other characters.
    """

    key = "fight"
    help_category = "combat"

    rules = COMBAT_RULES
    combat_handler_class = TBMagicTurnHandler


class CmdAttack(tb_basic.CmdAttack):
    """
    Attacks another character.

    Usage:
      attack <target>

    When in a fight, you may attack another character. The attack has
    a chance to hit, and if successful, will deal damage.
    """

    key = "attack"
    help_category = "combat"

    rules = COMBAT_RULES


class CmdPass(tb_basic.CmdPass):
    """
    Passes on your turn.

    Usage:
      pass

    When in a fight, you can use this command to end your turn early, even
    if there are still any actions you can take.
    """

    key = "pass"
    aliases = ["wait", "hold"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdDisengage(tb_basic.CmdDisengage):
    """
    Passes your turn and attempts to end combat.

    Usage:
      disengage

    Ends your turn early and signals that you're trying to end
    the fight. If all participants in a fight disengage, the
    fight ends.
    """

    key = "disengage"
    aliases = ["spare"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdLearnSpell(Command):
    """
    Learn a magic spell.

    Usage:
        learnspell <spell name>

    Adds a spell by name to your list of spells known.

    The following spells are provided as examples:

        |wmagic missile|n (3 MP): Fires three missiles that never miss. Can target
            up to three different enemies.

        |wflame shot|n (3 MP): Shoots a high-damage jet of flame at one target.

        |wcure wounds|n (5 MP): Heals damage on one target.

        |wmass cure wounds|n (10 MP): Like 'cure wounds', but can heal up to 5
            targets at once.

        |wfull heal|n (12 MP): Heals one target back to full HP.

        |wcactus conjuration|n (2 MP): Creates a cactus.
    """

    key = "learnspell"
    help_category = "magic"

    def func(self):
        """
        This performs the actual command.
        """
        spell_list = sorted(SPELLS.keys())
        args = self.args.lower()
        args = args.strip(" ")
        caller = self.caller
        spell_to_learn = []

        if not args or len(args) < 3:  # No spell given
            caller.msg("Usage: learnspell <spell name>")
            return

        for spell in spell_list:  # Match inputs to spells
            if args in spell.lower():
                spell_to_learn.append(spell)

        if spell_to_learn == []:  # No spells matched
            caller.msg("There is no spell with that name.")
            return
        if len(spell_to_learn) > 1:  # More than one match
            matched_spells = ", ".join(spell_to_learn)
            caller.msg("Which spell do you mean: %s?" % matched_spells)
            return

        if len(spell_to_learn) == 1:  # If one match, extract the string
            spell_to_learn = spell_to_learn[0]

        if spell_to_learn not in self.caller.db.spells_known:  # If the spell isn't known...
            caller.db.spells_known.append(spell_to_learn)  # ...then add the spell to the character
            caller.msg("You learn the spell '%s'!" % spell_to_learn)
            return
        if spell_to_learn in self.caller.db.spells_known:  # Already has the spell specified
            caller.msg("You already know the spell '%s'!" % spell_to_learn)
        """
        You will almost definitely want to replace this with your own system
        for learning spells, perhaps tied to character advancement or finding
        items in the game world that spells can be learned from.
        """


class CmdCast(MuxCommand):
    """
    Cast a magic spell that you know, provided you have the MP
    to spend on its casting.

    Usage:
        cast <spellname> [= <target1>, <target2>, etc...]

    Some spells can be cast on multiple targets, some can be cast
    on only yourself, and some don't need a target specified at all.
    Typing 'cast' by itself will give you a list of spells you know.
    """

    key = "cast"
    help_category = "magic"

    rules = COMBAT_RULES

    def func(self):
        """
        This performs the actual command.

        Note: This is a quite long command, since it has to cope with all
        the different circumstances in which you may or may not be able
        to cast a spell. None of the spell's effects are handled by the
        command - all the command does is verify that the player's input
        is valid for the spell being cast and then call the spell's
        function.
        """
        caller = self.caller

        if not self.lhs or len(self.lhs) < 3:  # No spell name given
            caller.msg("Usage: cast <spell name> = <target>, <target2>, ...")
            if not caller.db.spells_known:
                caller.msg("You don't know any spells.")
                return
            else:
                caller.db.spells_known = sorted(caller.db.spells_known)
                spells_known_msg = "You know the following spells:|/" + "|/".join(
                    caller.db.spells_known
                )
                caller.msg(spells_known_msg)  # List the spells the player knows
                return

        spellname = self.lhs.lower()  # noqa  - not used but potentially useful
        spell_to_cast = []
        spell_targets = []

        if not self.rhs:
            spell_targets = []
        elif self.rhs.lower() in ["me", "self", "myself"]:
            spell_targets = [caller]
        elif len(self.rhs) > 2:
            spell_targets = self.rhslist

        for spell in caller.db.spells_known:  # Match inputs to spells
            if self.lhs in spell.lower():
                spell_to_cast.append(spell)

        if spell_to_cast == []:  # No spells matched
            caller.msg("You don't know a spell of that name.")
            return
        if len(spell_to_cast) > 1:  # More than one match
            matched_spells = ", ".join(spell_to_cast)
            caller.msg("Which spell do you mean: %s?" % matched_spells)
            return

        if len(spell_to_cast) == 1:  # If one match, extract the string
            spell_to_cast = spell_to_cast[0]

        if spell_to_cast not in SPELLS:  # Spell isn't defined
            caller.msg("ERROR: Spell %s is undefined" % spell_to_cast)
            return

        # Time to extract some info from the chosen spell!
        spelldata = SPELLS[spell_to_cast]

        # Add in some default data if optional parameters aren't specified
        if "combat_spell" not in spelldata:
            spelldata.update({"combat_spell": True})
        if "noncombat_spell" not in spelldata:
            spelldata.update({"noncombat_spell": True})
        if "max_targets" not in spelldata:
            spelldata.update({"max_targets": 1})

        # Store any superfluous options as kwargs to pass to the spell function
        kwargs = {}
        spelldata_opts = [
            "spellfunc",
            "target",
            "cost",
            "combat_spell",
            "noncombat_spell",
            "max_targets",
        ]
        for key in spelldata:
            if key not in spelldata_opts:
                kwargs.update({key: spelldata[key]})

        # If caster doesn't have enough MP to cover the spell's cost, give error and return
        if spelldata["cost"] > caller.db.mp:
            caller.msg("You don't have enough MP to cast '%s'." % spell_to_cast)
            return

        # If in combat and the spell isn't a combat spell, give error message and return
        if spelldata["combat_spell"] is False and self.rules.is_in_combat(caller):
            caller.msg("You can't use the spell '%s' in combat." % spell_to_cast)
            return

        # If not in combat and the spell isn't a non-combat spell, error ms and return.
        if spelldata["noncombat_spell"] is False and self.rules.is_in_combat(caller) is False:
            caller.msg("You can't use the spell '%s' outside of combat." % spell_to_cast)
            return

        # If spell takes no targets and one is given, give error message and return
        if len(spell_targets) > 0 and spelldata["target"] == "none":
            caller.msg("The spell '%s' isn't cast on a target." % spell_to_cast)
            return

        # If no target is given and spell requires a target, give error message
        if spelldata["target"] not in ["self", "none"]:
            if len(spell_targets) == 0:
                caller.msg("The spell '%s' requires a target." % spell_to_cast)
                return

        # If more targets given than maximum, give error message
        if len(spell_targets) > spelldata["max_targets"]:
            targplural = "target"
            if spelldata["max_targets"] > 1:
                targplural = "targets"
            caller.msg(
                "The spell '%s' can only be cast on %i %s."
                % (spell_to_cast, spelldata["max_targets"], targplural)
            )
            return

        # Set up our candidates for targets
        target_candidates = []

        # If spell targets 'any' or 'other', any object in caster's inventory or location
        # can be targeted by the spell.
        if spelldata["target"] in ["any", "other"]:
            target_candidates = caller.location.contents + caller.contents

        # If spell targets 'anyobj', only non-character objects can be targeted.
        if spelldata["target"] == "anyobj":
            prefilter_candidates = caller.location.contents + caller.contents
            for thing in prefilter_candidates:
                if not thing.attributes.has("max_hp"):  # Has no max HP, isn't a fighter
                    target_candidates.append(thing)

        # If spell targets 'anychar' or 'otherchar', only characters can be targeted.
        if spelldata["target"] in ["anychar", "otherchar"]:
            prefilter_candidates = caller.location.contents
            for thing in prefilter_candidates:
                if thing.attributes.has("max_hp"):  # Has max HP, is a fighter
                    target_candidates.append(thing)

        # Now, match each entry in spell_targets to an object in the search candidates
        matched_targets = []
        for target in spell_targets:
            match = caller.search(target, candidates=target_candidates)
            matched_targets.append(match)
        spell_targets = matched_targets

        # If no target is given and the spell's target is 'self', set target to self
        if len(spell_targets) == 0 and spelldata["target"] == "self":
            spell_targets = [caller]

        # Give error message if trying to cast an "other" target spell on yourself
        if spelldata["target"] in ["other", "otherchar"]:
            if caller in spell_targets:
                caller.msg("You can't cast '%s' on yourself." % spell_to_cast)
                return

        # Return if "None" in target list, indicating failed match
        if None in spell_targets:
            # No need to give an error message, as 'search' gives one by default.
            return

        # Give error message if repeats in target list
        if len(spell_targets) != len(set(spell_targets)):
            caller.msg("You can't specify the same target more than once!")
            return

        # Finally, we can cast the spell itself. Note that MP is not deducted here!
        try:
            spelldata["spellfunc"](
                caller, spell_to_cast, spell_targets, spelldata["cost"], **kwargs
            )
        except Exception:
            log_trace("Error in callback for spell: %s." % spell_to_cast)


class CmdRest(Command):
    """
    Recovers damage and restores MP.

    Usage:
      rest

    Resting recovers your HP and MP to their maximum, but you can
    only rest if you're not in a fight.
    """

    key = "rest"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        "This performs the actual command."

        if self.rules.is_in_combat(self.caller):  # If you're in combat
            self.caller.msg("You can't rest while you're in combat.")
            return

        self.caller.db.hp = self.caller.db.max_hp  # Set current HP to maximum
        self.caller.db.mp = self.caller.db.max_mp  # Set current MP to maximum
        self.caller.location.msg_contents("%s rests to recover HP and MP." % self.caller)
        # You'll probably want to replace this with your own system for recovering HP and MP.


class CmdStatus(Command):
    """
    Gives combat information.

    Usage:
      status

    Shows your current and maximum HP and your distance from
    other targets in combat.
    """

    key = "status"
    help_category = "combat"

    def func(self):
        "This performs the actual command."
        char = self.caller

        if not char.db.max_hp:  # Character not initialized, IE in unit tests
            char.db.hp = 100
            char.db.max_hp = 100
            char.db.spells_known = []
            char.db.max_mp = 20
            char.db.mp = char.db.max_mp

        char.msg(
            "You have %i / %i HP and %i / %i MP."
            % (char.db.hp, char.db.max_hp, char.db.mp, char.db.max_mp)
        )


class CmdCombatHelp(tb_basic.CmdCombatHelp):
    """
    View help or a list of topics

    Usage:
      help <topic or command>
      help list
      help all

    This will search for help on commands and other
    topics related to the game.
    """


class BattleCmdSet(default_cmds.CharacterCmdSet):
    """
    This command set includes all the commmands used in the battle system.
    """

    key = "DefaultCharacter"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        self.add(CmdFight())
        self.add(CmdAttack())
        self.add(CmdRest())
        self.add(CmdPass())
        self.add(CmdDisengage())
        self.add(CmdCombatHelp())
        self.add(CmdLearnSpell())
        self.add(CmdCast())
        self.add(CmdStatus())



"""
A simple quest system for EvAdventure.

A quest is represented by a quest-handler sitting as
.quest on a Character. Individual Quests are objects
that track the state and can have multiple steps, each
of which are checked off during the quest's progress.

The player can use the quest handler to track the
progress of their quests.

A quest ending can mean a reward or the start of
another quest.

"""

from copy import copy, deepcopy

from evennia.utils import dbserialize


class EvAdventureQuest:
    """
    This represents a single questing unit of quest.

    Properties:
        name (str): Main identifier for the quest.
        category (str, optional): This + name must be globally unique.
    it ends - it then pauses after the last completed step.

    Each step of the quest is represented by a `.step_<stepname>` method. This should check
    the status of the quest-step and update the `.current_step` or call `.complete()`. There
    are also `.help_<stepname>` which is either a class-level help string or a method
    returning a help text. All properties should be stored on the quester.

    Example:
    ```py
    class MyQuest(EvAdventureQuest):
        '''A quest with two steps that ar'''

        start_step = "A"

        help_A = "You need a '_quest_A_flag' on yourself to finish this step!"
        help_B = "Finally, you need more than 4 items in your inventory!"

        def step_A(self, *args, **kwargs):
            if self.quester.db._quest_A_flag == True:
                self.quester.msg("Completed the first step of the quest.")
                self.current_step = "end"
                self.progress()

        def step_end(self, *args, **kwargs):
            if len(self.quester.contents) > 4:
                self.quester.msg("Quest complete!")
                self.complete()
    ```
    """

    key = "basequest"
    desc = "This is the base quest class"
    start_step = "start"

    completed_text = "This quest is completed!"
    abandoned_text = "This quest is abandoned."

    # help entries for quests (could also be methods)
    help_start = "You need to start first"
    help_end = "You need to end the quest"

    def __init__(self, quester, start_step=None):
        if " " in self.key:
            raise TypeError("The Quest name must not have spaces in it.")

        self.quester = quester
        self._current_step = start_step or self.start_step
        self.is_completed = False
        self.is_abandoned = False

    def __serialize_dbobjs__(self):
        self.quester = dbserialize.dbserialize(self.quester)

    def __deserialize_dbobjs__(self):
        if isinstance(self.quester, bytes):
            self.quester = dbserialize.dbunserialize(self.quester)

    @property
    def questhandler(self):
        return self.quester.quests

    @property
    def current_step(self):
        return self._current_step

    @current_step.setter
    def current_step(self, step_name):
        self._current_step = step_name
        self.questhandler.do_save = True

    def abandon(self):
        """
        Call when quest is abandoned.

        """
        self.is_abandoned = True
        self.cleanup()

    def complete(self):
        """
        Call this to end the quest.

        """
        self.is_completed = True
        self.cleanup()

    def progress(self, *args, **kwargs):
        """
        This is called whenever the environment expects a quest may need stepping. This will
        determine which quest-step we are on and run `step_<stepname>`, which in turn will figure
        out if the step is complete or not.

        Args:
            *args, **kwargs: Will be passed into the step method.

        """
        if not (self.is_completed or self.is_abandoned):
            getattr(self, f"step_{self.current_step}")(*args, **kwargs)

    def help(self):
        """
        This is used to get help (or a reminder) of what needs to be done to complete the current
        quest-step.

        Returns:
            str: The help text for the current step.

        """
        if self.is_completed:
            return self.completed_text
        if self.is_abandoned:
            return self.abandoned_text

        help_resource = (
            getattr(self, f"help_{self.current_step}", None)
            or "You need to {self.current_step} ..."
        )
        if callable(help_resource):
            # the help_<current_step> can be a method to call
            return help_resource()
        else:
            # normally it's just a string
            return str(help_resource)

    # step methods and hooks

    def step_start(self, *args, **kwargs):
        """
        Example step that completes immediately.

        """
        self.complete()

    def cleanup(self):
        """
        This is called both when completing the quest, or when it is abandoned prematurely.
        Make sure to cleanup any quest-related data stored when following the quest.

        """
        pass


class EvAdventureQuestHandler:
    """
    This sits on the Character, as `.quests`.

    It's initiated using a lazy property on the Character:

    ```
    @lazy_property
    def quests(self):
        return EvAdventureQuestHandler(self)
    ```

    """

    quest_storage_attribute_key = "_quests"
    quest_storage_attribute_category = "evadventure"

    def __init__(self, obj):
        self.obj = obj
        self.do_save = False
        self._load()

    def _load(self):
        self.storage = self.obj.attributes.get(
            self.quest_storage_attribute_key,
            category=self.quest_storage_attribute_category,
            default={},
        )

    def _save(self):
        self.obj.attributes.add(
            self.quest_storage_attribute_key,
            self.storage,
            category=self.quest_storage_attribute_category,
        )
        self._load()  # important
        self.do_save = False

    def has(self, quest_key):
        """
        Check if a given quest is registered with the Character.

        Args:
            quest_key (str): The name of the quest to check for.
            quest_category (str, optional): Quest category, if any.

        Returns:
            bool: If the character is following this quest or not.

        """
        return bool(self.storage.get(quest_key))

    def get(self, quest_key):
        """
        Get the quest stored on character, if any.

        Args:
            quest_key (str): The name of the quest to check for.

        Returns:
            EvAdventureQuest or None: The quest stored, or None if
                Character is not on this quest.

        """
        return self.storage.get(quest_key)

    def add(self, quest):
        """
        Add a new quest

        Args:
            quest (EvAdventureQuest): The quest class to start.

        """
        self.storage[quest.key] = quest(self.obj)
        self._save()

    def remove(self, quest_key):
        """
        Remove a quest. If not complete, it will be abandoned.

        Args:
            quest_key (str): The quest to remove.

        """
        quest = self.storage.pop(quest_key, None)
        if not quest.is_completed:
            # make sure to cleanup
            quest.abandon()
        self._save()

    def get_help(self, quest_key=None):
        """
        Get help text for a quest or for all quests. The help text is
        a combination of the description of the quest and the help-text
        of the current step.

        Args:
            quest_key (str, optional): The quest-key. If not given, get help for all
                quests in handler.

        Returns:
            list: Help texts, one for each quest, or only one if `quest_key` is given.

        """
        help_texts = []
        if quest_key in self.storage:
            quests = [self.storage[quest_key]]
        else:
            quests = self.storage.values()

        for quest in quests:
            help_texts.append(f"|c{quest.key}|n\n {quest.desc}\n\n - {quest.help()}")
        return help_texts

    def progress(self, quest_key=None, *args, **kwargs):
        """
        Check progress of a given quest or all quests.

        Args:
            quest_key (str, optional): If given, check the progress of this quest (if we have it),
                otherwise check progress on all quests.
            *args, **kwargs: Will be passed into each quest's `progress` call.

        """
        if quest_key in self.storage:
            quests = [self.storage[quest_key]]
        else:
            quests = self.storage.values()

        for quest in quests:
            quest.progress(*args, **kwargs)

        if self.do_save:
            # do_save is set by the quest
            self._save()



"""
Unit tests for the scripts package

"""

from collections import defaultdict
from unittest import TestCase, mock

from parameterized import parameterized

from evennia import DefaultScript
from evennia.objects.objects import DefaultObject
from evennia.scripts.manager import ScriptDBManager
from evennia.scripts.models import ObjectDoesNotExist, ScriptDB
from evennia.scripts.monitorhandler import MonitorHandler
from evennia.scripts.ondemandhandler import OnDemandHandler, OnDemandTask
from evennia.scripts.scripts import DoNothing, ExtendedLoopingCall
from evennia.scripts.tickerhandler import TickerHandler
from evennia.utils.create import create_script
from evennia.utils.dbserialize import dbserialize
from evennia.utils.test_resources import BaseEvenniaTest, EvenniaTest


class TestScript(BaseEvenniaTest):
    def test_create(self):
        "Check the script can be created via the convenience method."
        with mock.patch("evennia.scripts.scripts.DefaultScript.at_init") as mockinit:
            obj, errors = DefaultScript.create("useless-machine")
            self.assertTrue(obj, errors)
            self.assertFalse(errors, errors)
            mockinit.assert_called()


class TestTickerHandler(TestCase):
    """Test the TickerHandler class"""

    def test_store_key_raises_RunTimeError(self):
        """Test _store_key method raises RuntimeError for interval < 1"""
        with self.assertRaises(RuntimeError):
            th = TickerHandler()
            th._store_key(None, None, 0, None)

    def test_remove_raises_RunTimeError(self):
        """Test remove method raises RuntimeError for catching old ordering of arguments"""
        with self.assertRaises(RuntimeError):
            th = TickerHandler()
            th.remove(callback=1)


class TestScriptDBManager(TestCase):
    """Test the ScriptDBManger class"""

    def test_not_obj_return_empty_list(self):
        """Test get_all_scripts_on_obj returns empty list for falsy object"""
        manager_obj = ScriptDBManager()
        returned_list = manager_obj.get_all_scripts_on_obj(False)
        self.assertEqual(returned_list, [])


class TestingListIntervalScript(DefaultScript):
    """
    A script that does nothing. Used to test listing of script with nonzero intervals.
    """

    def at_script_creation(self):
        """
        Setup the script
        """
        self.key = "interval_test"
        self.desc = "This is an empty placeholder script."
        self.interval = 1
        self.repeats = 1


class TestScriptHandler(BaseEvenniaTest):
    """
    Test the ScriptHandler class.

    """

    def setUp(self):
        self.obj, self.errors = DefaultObject.create("test_object")
        self.obj.scripts.add(TestingListIntervalScript)

    def tearDown(self):
        self.obj.delete()

    def test_start(self):
        "Check that ScriptHandler start function works correctly"
        self.num = self.obj.scripts.start(self.obj.scripts.all()[0].key)
        self.assertEqual(self.num, 1)

    def test_list_script_intervals(self):
        "Checks that Scripthandler __str__ function lists script intervals correctly"
        self.str = str(self.obj.scripts)
        self.assertTrue("None/1" in self.str)
        self.assertTrue("1 repeats" in self.str)

    def test_get_all_scripts(self):
        "Checks that Scripthandler get_all returns correct number of scripts"
        self.assertEqual([script.key for script in self.obj.scripts.all()], ["interval_test"])

    def test_get_script(self):
        "Checks that Scripthandler get function returns correct script"
        script = self.obj.scripts.get("interval_test")
        self.assertTrue(bool(script))


class TestScriptDB(TestCase):
    "Check the singleton/static ScriptDB object works correctly"

    def setUp(self):
        self.scr = create_script(DoNothing)

    def tearDown(self):
        try:
            self.scr.delete()
        except ObjectDoesNotExist:
            pass
        del self.scr

    def test_delete(self):
        "Check the script is removed from the database"
        self.scr.delete()
        self.assertFalse(self.scr in ScriptDB.objects.get_all_scripts())

    def test_double_delete(self):
        "What should happen? Isn't it already deleted?"
        with self.assertRaises(ObjectDoesNotExist):
            self.scr.delete()
            self.scr.delete()

    def test_deleted_script_fails_start(self):
        "Would it ever be necessary to start a deleted script?"
        self.scr.delete()
        with self.assertRaises(ScriptDB.DoesNotExist):  # See issue #509
            self.scr.start()
        # Check the script is not recreated as a side-effect
        self.assertFalse(self.scr in ScriptDB.objects.get_all_scripts())


class TestExtendedLoopingCall(TestCase):
    """
    Test the ExtendedLoopingCall class.

    """

    @mock.patch("evennia.scripts.scripts.LoopingCall")
    def test_start__nodelay(self, MockClass):
        """Test the .start method with no delay"""

        callback = mock.MagicMock()
        loopcall = ExtendedLoopingCall(callback)
        loopcall.__call__ = mock.MagicMock()
        loopcall._scheduleFrom = mock.MagicMock()
        loopcall.clock.seconds = mock.MagicMock(return_value=0)

        loopcall.start(20, now=True, start_delay=None, count_start=1)
        loopcall._scheduleFrom.assert_not_called()

    @mock.patch("evennia.scripts.scripts.LoopingCall")
    def test_start__delay(self, MockLoopingCall):
        """Test the .start method with delay"""

        callback = mock.MagicMock()
        MockLoopingCall.clock.seconds = mock.MagicMock(return_value=0)

        loopcall = ExtendedLoopingCall(callback)
        loopcall.__call__ = mock.MagicMock()
        loopcall.clock.seconds = mock.MagicMock(return_value=121)
        loopcall._scheduleFrom = mock.MagicMock()

        loopcall.start(20, now=False, start_delay=10, count_start=1)

        loopcall.__call__.assert_not_called()
        self.assertEqual(loopcall.interval, 20)
        loopcall._scheduleFrom.assert_called_with(121)

    def test_start_invalid_interval(self):
        """Test the .start method with interval less than zero"""
        with self.assertRaises(ValueError):
            callback = mock.MagicMock()
            loopcall = ExtendedLoopingCall(callback)
            loopcall.start(-1, now=True, start_delay=None, count_start=1)

    def test__call__when_delay(self):
        """Test __call__ modifies start_delay and starttime if start_delay was previously set"""
        callback = mock.MagicMock()
        loopcall = ExtendedLoopingCall(callback)
        loopcall.clock.seconds = mock.MagicMock(return_value=1)
        loopcall.start_delay = 2
        loopcall.starttime = 0

        loopcall()

        self.assertEqual(loopcall.start_delay, None)
        self.assertEqual(loopcall.starttime, 1)

    def test_force_repeat(self):
        """Test forcing script to run that is scheduled to run in the future"""
        callback = mock.MagicMock()
        loopcall = ExtendedLoopingCall(callback)
        loopcall.clock.seconds = mock.MagicMock(return_value=0)

        loopcall.start(20, now=False, start_delay=5, count_start=0)
        loopcall.force_repeat()
        loopcall.stop()

        callback.assert_called_once()


def dummy_func():
    """Dummy function used as callback parameter"""
    return 0


class TestMonitorHandler(TestCase):
    """
    Test the MonitorHandler class.
    """

    def setUp(self):
        self.handler = MonitorHandler()

    def test_add(self):
        """Tests that adding an object to the monitor handler works correctly"""
        obj = mock.Mock()
        fieldname = "db_add"
        callback = dummy_func
        idstring = "test"

        self.handler.add(obj, fieldname, callback, idstring=idstring)

        self.assertIn(fieldname, self.handler.monitors[obj])
        self.assertIn(idstring, self.handler.monitors[obj][fieldname])
        self.assertEqual(self.handler.monitors[obj][fieldname][idstring], (callback, False, {}))

    def test_remove(self):
        """Tests that removing an object from the monitor handler works correctly"""
        obj = mock.Mock()
        fieldname = "db_remove"
        callback = dummy_func
        idstring = "test_remove"

        """Add an object to the monitor handler and then remove it"""
        self.handler.add(obj, fieldname, callback, idstring=idstring)
        self.handler.remove(obj, fieldname, idstring=idstring)
        self.assertEquals(self.handler.monitors[obj][fieldname], {})

    def test_add_with_invalid_function(self):
        obj = mock.Mock()
        """Tests that add method rejects objects where callback is not a function"""
        fieldname = "db_key"
        callback = "not_a_function"

        self.handler.add(obj, fieldname, callback)
        self.assertNotIn(fieldname, self.handler.monitors[obj])

    def test_all(self):
        """Tests that all method correctly returns information about added objects"""
        obj = [mock.Mock(), mock.Mock()]
        fieldname = ["db_all1", "db_all2"]
        callback = dummy_func
        idstring = ["test_all1", "test_all2"]

        self.handler.add(obj[0], fieldname[0], callback, idstring=idstring[0])
        self.handler.add(obj[1], fieldname[1], callback, idstring=idstring[1], persistent=True)

        output = self.handler.all()
        self.assertEquals(
            output,
            [
                (obj[0], fieldname[0], idstring[0], False, {}),
                (obj[1], fieldname[1], idstring[1], True, {}),
            ],
        )

    def test_clear(self):
        """Tests that the clear function correctly clears the monitor handler"""
        obj = mock.Mock()
        fieldname = "db_add"
        callback = dummy_func
        idstring = "test"

        self.handler.add(obj, fieldname, callback, idstring=idstring)
        self.assertIn(obj, self.handler.monitors)

        self.handler.clear()
        self.assertNotIn(obj, self.handler.monitors)
        self.assertEquals(defaultdict(lambda: defaultdict(dict)), self.handler.monitors)

    def test_add_remove_attribute(self):
        """Tests that adding and removing an object attribute to the monitor handler works correctly"""
        obj = mock.Mock()
        obj.name = "testaddattribute"
        fieldname = "name"
        callback = dummy_func
        idstring = "test"
        category = "testattribute"

        """Add attribute to handler and assert that it has been added"""
        self.handler.add(obj, fieldname, callback, idstring=idstring, category=category)

        index = obj.attributes.get(fieldname, return_obj=True)
        name = "db_value[testattribute]"

        self.assertIn(name, self.handler.monitors[index])
        self.assertIn(idstring, self.handler.monitors[index][name])
        self.assertEqual(self.handler.monitors[index][name][idstring], (callback, False, {}))

        """Remove attribute from the handler and assert that it is gone"""
        self.handler.remove(obj, fieldname, idstring=idstring, category=category)
        self.assertEquals(self.handler.monitors[index][name], {})


class TestOnDemandTask(EvenniaTest):
    """
    Test the OnDemandTask class.

    """

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_no_stages__no_autostart(self, mock_runtime):
        mock_runtime.return_value = 1000
        task = OnDemandTask("rose", "flower", autostart=False)

        self.assertEqual(task.key, "rose")
        self.assertEqual(task.category, "flower")
        self.assertEqual(task.start_time, None)

        self.assertEqual(str(task), "OnDemandTask(rose[flower] (dt=0s), stage=None)")
        self.assertEqual(task.get_dt(), 0)
        self.assertEqual(task.start_time, 1000)

        mock_runtime.return_value = 3000

        self.assertEqual(task.get_dt(), 2000)
        self.assertEqual(task.get_stage(), None)
        self.assertEqual(task.start_time, 1000)

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_stages_autostart(self, mock_runtime):
        START_TIME = 1000
        mock_runtime.return_value = START_TIME
        task = OnDemandTask(
            "rose",
            "flower",
            stages={0: "seedling", 100: "bud", 200: "flower", 300: "wilted", 400: "dead"},
        )
        self.assertEqual(task.start_time, 1000)
        self.assertEqual(
            task.stages,
            {
                0: ("seedling", None),
                100: ("bud", None),
                200: ("flower", None),
                300: ("wilted", None),
                400: ("dead", None),
            },
        )

        # step through the stages
        self.assertEqual(task.get_dt(), 0)

        self.assertEqual(task.get_stage(), "seedling")

        mock_runtime.return_value = START_TIME + 20
        self.assertEqual(task.get_stage(), "seedling")

        mock_runtime.return_value = START_TIME + 99.99
        self.assertEqual(task.get_stage(), "seedling")

        mock_runtime.return_value = START_TIME + 100
        self.assertEqual(task.get_stage(), "bud")

        mock_runtime.return_value = START_TIME + 250.14
        self.assertEqual(task.get_stage(), "flower")

        mock_runtime.return_value = START_TIME + 300
        self.assertEqual(task.get_stage(), "wilted")

        mock_runtime.return_value = START_TIME + 400.0
        self.assertEqual(task.get_stage(), "dead")

        mock_runtime.return_value = START_TIME + 10000
        self.assertEqual(task.get_stage(), "dead")

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_stagefuncs(self, mock_runtime):
        START_TIME = 0
        mock_runtime.return_value = START_TIME

        def statefunc(task):
            task.start_time = 2000

        task = OnDemandTask(
            "rose",
            "flower",
            stages={
                0: "seedling",
                100: "bud",
                200: "flower",
                300: "wilted",
                400: ("dead", statefunc),
            },
        )

        self.assertEqual(task.get_stage(), "seedling")
        mock_runtime.return_value = START_TIME + 400
        self.assertEqual(task.get_stage(), "dead")
        self.assertEqual(task.start_time, 2000)

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_stagefunc_loop(self, mock_runtime):
        START_TIME = 0
        mock_runtime.return_value = START_TIME

        task = OnDemandTask(
            "rose",
            "flower",
            stages={
                0: "seedling",
                50: "bud",
                150: "flower",
                300: "wilted",
                400: "dead",
                500: ("_loop", OnDemandTask.stagefunc_loop),
            },
        )

        self.assertAlmostEqual(task.get_dt(), 0)
        self.assertEqual(task.get_stage(), "seedling")

        mock_runtime.return_value = START_TIME + 500
        self.assertEqual(task.get_dt(), 0)
        self.assertEqual(task.get_stage(), "seedling")

        mock_runtime.return_value = START_TIME + 600
        self.assertEqual(task.get_dt(), 100)
        self.assertEqual(task.iterations, 1)
        self.assertEqual(task.get_stage(), "bud")

        # wait a long time, should loop back indefinitely, counting iterations
        mock_runtime.return_value = START_TIME + 10250
        self.assertEqual(task.get_dt(), 250)
        self.assertEqual(task.iterations, 20)
        self.assertEqual(task.get_stage(), "flower")

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_stagefunc_bounce(self, mock_runtime):
        START_TIME = 0
        mock_runtime.return_value = START_TIME

        task = OnDemandTask(
            "reactor",
            "nuclear",
            stages={
                0: ("cold", OnDemandTask.stagefunc_bounce),
                50: "lukewarm",
                150: "warm",
                300: "hot",
                400: ("HOT!", OnDemandTask.stagefunc_bounce),
            },
        )

        self.assertAlmostEqual(task.get_dt(), 0)
        self.assertEqual(task.get_stage(), "cold")

        mock_runtime.return_value = START_TIME + 400
        self.assertEqual(task.get_dt(), 0)
        self.assertEqual(task.get_stage(), "HOT!")
        self.assertEqual(task.iterations, 1)

        # we should be going back down the sequence
        mock_runtime.return_value = START_TIME + 450
        self.assertEqual(task.get_dt(), 50)
        self.assertEqual(task.get_stage(), "HOT!")

        mock_runtime.return_value = START_TIME + 500
        self.assertEqual(task.get_dt(), 100)
        self.assertEqual(task.get_stage(), "hot")

        mock_runtime.return_value = START_TIME + 650
        self.assertEqual(task.get_dt(), 250)
        self.assertEqual(task.get_stage(), "warm")

        mock_runtime.return_value = START_TIME + 750
        self.assertEqual(task.get_dt(), 350)
        self.assertEqual(task.get_stage(), "lukewarm")

        mock_runtime.return_value = START_TIME + 800
        self.assertEqual(task.get_dt(), 0)
        self.assertEqual(task.iterations, 2)
        self.assertEqual(task.get_stage(), "cold")

        # back up again
        mock_runtime.return_value = START_TIME + 950
        self.assertEqual(task.get_dt(), 150)
        self.assertEqual(task.iterations, 2)
        self.assertEqual(task.get_stage(), "warm")

        # Waiting a long time
        mock_runtime.return_value = START_TIME + 10250
        self.assertEqual(task.get_dt(), 250)
        self.assertEqual(task.iterations, 25)
        self.assertEqual(task.get_stage(), "warm")


class TestOnDemandHandler(EvenniaTest):
    """
    Test the OnDemandHandler class.

    """

    def setUp(self):
        super(TestOnDemandHandler, self).setUp()
        self.handler = OnDemandHandler()
        self.task1 = OnDemandTask(
            "rose",
            "flower",
            stages={0: "seedling", 100: "bud", 200: "flower", 300: "wilted", 400: "dead"},
        )
        self.task2 = OnDemandTask(
            "daffodil",
            "flower",
            stages={0: "seedling", 50: "bud", 100: "flower", 150: "wilted", 200: "dead"},
        )
        self.task3 = OnDemandTask("test", None)

    def test_add_get(self):
        self.handler.add("rose", category="flower", stages={0: "seedling"})
        task = self.handler.get("rose", "flower")
        self.assertEqual(
            (task.key, task.category, task.stages), ("rose", "flower", {0: ("seedling", None)})
        )
        self.assertEqual(self.handler.get("rose"), None)  # no category

    def test_batch_add(self):
        self.handler.batch_add(self.task1, self.task2, self.task3)
        task1 = self.handler.get("rose", "flower")
        task2 = self.handler.get("daffodil", "flower")
        task3 = self.handler.get("test")
        self.assertEqual((task1.key, task1.category), ("rose", "flower"))
        self.assertEqual((task2.key, task2.category), ("daffodil", "flower"))
        self.assertEqual((task3.key, task3.category), ("test", None))

    def test_remove(self):
        self.handler.add(self.task1)
        self.handler.add(self.task2)
        self.handler.remove(self.task1)
        self.assertEqual(self.handler.get("rose", "flower"), None)
        self.assertEqual(self.handler.get("daffodil", "flower"), self.task2)

    def test_batch_remove(self):
        self.handler.batch_add(self.task1, self.task2, self.task3)
        self.handler.batch_remove(self.task1, self.task2)
        self.assertEqual(self.handler.get("rose", "flower"), None)
        self.assertEqual(self.handler.get("daffodil", "flower"), None)
        self.assertEqual(self.handler.get("test"), self.task3)

    def test_all(self):
        self.handler.batch_add(self.task1, self.task2, self.task3)
        self.assertEqual(
            self.handler.all(),
            {
                (self.task1.key, self.task1.category): self.task1,
                (self.task2.key, self.task2.category): self.task2,
                (self.task3.key, self.task3.category): self.task3,
            },
        )

    def test_clear(self):
        self.handler.batch_add(self.task1, self.task2, self.task3)
        self.handler.clear(all_on_none=False)  # only task3 gone (None
        self.assertEqual(
            self.handler.all(),
            {
                (self.task1.key, self.task1.category): self.task1,
                (self.task2.key, self.task2.category): self.task2,
            },
        )
        self.handler.clear()  # all gone
        self.assertEqual(self.handler.all(), {})

    def test_save_and_load(self):
        self.handler.add(self.task1)
        self.handler.add(self.task2)
        self.handler.save()
        self.handler.load()
        task1 = self.handler.get("rose", "flower")
        task2 = self.handler.get(self.task2)
        self.assertEqual((task1.key, task1.category), ("rose", "flower"))
        self.assertEqual((task2.key, task2.category), ("daffodil", "flower"))

    @mock.patch("evennia.scripts.ondemandhandler.OnDemandTask.runtime")
    def test_get_dt_and_stage(self, mock_runtime):
        START_TIME = 0

        mock_runtime.return_value = START_TIME
        self.handler.batch_add(self.task1, self.task2)

        for task in self.handler.tasks.values():
            task.start_time = START_TIME

        self.assertEqual(self.handler.get_dt("rose", "flower"), 0)
        self.assertEqual(self.handler.get_dt("daffodil", "flower"), 0)

        mock_runtime.return_value = START_TIME + 50
        self.assertEqual(self.handler.get_dt("rose", "flower"), 50)
        self.assertEqual(self.handler.get_dt("daffodil", "flower"), 50)
        self.assertEqual(self.handler.get_stage("rose", "flower"), "seedling")
        self.assertEqual(self.handler.get_stage("daffodil", "flower"), "bud")

        mock_runtime.return_value = START_TIME + 150
        self.assertEqual(self.handler.get_dt("rose", "flower"), 150)
        self.assertEqual(self.handler.get_dt("daffodil", "flower"), 150)
        self.assertEqual(self.handler.get_stage("rose", "flower"), "bud")
        self.assertEqual(self.handler.get_stage("daffodil", "flower"), "wilted")

        mock_runtime.return_value = START_TIME + 250
        self.assertEqual(self.handler.get_dt("rose", "flower"), 250)
        self.assertEqual(self.handler.get_dt("daffodil", "flower"), 250)
        self.assertEqual(self.handler.get_stage("rose", "flower"), "flower")
        self.assertEqual(self.handler.get_stage("daffodil", "flower"), "dead")

        mock_runtime.return_value = START_TIME + 10000
        self.assertEqual(self.handler.get_dt("rose", "flower"), 10000)
        self.assertEqual(self.handler.get_dt("daffodil", "flower"), 10000)
        self.assertEqual(self.handler.get_stage("rose", "flower"), "dead")
        self.assertEqual(self.handler.get_stage("daffodil", "flower"), "dead")



"""
Tests for EvTable component.

"""

from unittest import skip

from evennia.utils import ansi, evtable
from evennia.utils.test_resources import EvenniaTestCase


class TestEvTable(EvenniaTestCase):
    def _validate(self, expected, result):
        """easier debug"""
        expected = ansi.strip_ansi(expected).strip()
        result = ansi.strip_ansi(result).strip()

        err = f"\n{'expected':-^60}\n{expected}\n{'result':-^60}\n{result}\n{'':-^60}"
        self.assertEqual(expected, result, err)

    def test_base(self):
        """
        Create plain table.

        """
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )

        expected = """
+----------+----------+----------+
| Heading1 | Heading2 | Heading3 |
+~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~+
| 1        | 4        | 7        |
+----------+----------+----------+
| 2        | 5        | 8        |
+----------+----------+----------+
| 3        | 6        | 9        |
+----------+----------+----------+
"""

        self._validate(expected, str(table))

    def test_table_with_short_header(self):
        """
        Don't provide header3

        """
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )
        expected = """
+----------+----------+---+
| Heading1 | Heading2 |   |
+~~~~~~~~~~+~~~~~~~~~~+~~~+
| 1        | 4        | 7 |
+----------+----------+---+
| 2        | 5        | 8 |
+----------+----------+---+
| 3        | 6        | 9 |
+----------+----------+---+
"""

        self._validate(expected, str(table))

    def test_add_column(self):
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )
        table.add_column("|rThis is long data|n", "|bThis is even longer data|n")

        expected = """
+----------+----------+----------+--------------------------+
| Heading1 | Heading2 | Heading3 |                          |
+~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~~~~~~~~~~~~~~~~~+
| 1        | 4        | 7        | This is long data        |
+----------+----------+----------+--------------------------+
| 2        | 5        | 8        | This is even longer data |
+----------+----------+----------+--------------------------+
| 3        | 6        | 9        |                          |
+----------+----------+----------+--------------------------+
"""
        self._validate(expected, str(table))

    def test_add_row(self):
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )
        table.add_row("This is a single row")

        expected = """
+----------------------+----------+----------+
| Heading1             | Heading2 | Heading3 |
+~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~+
| 1                    | 4        | 7        |
+----------------------+----------+----------+
| 2                    | 5        | 8        |
+----------------------+----------+----------+
| 3                    | 6        | 9        |
+----------------------+----------+----------+
| This is a single row |          |          |
+----------------------+----------+----------+
"""
        self._validate(expected, str(table))

    def test_right_align(self):
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="r",
        )

        expected = """
+----------+----------+----------+
| Heading1 | Heading2 | Heading3 |
+~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~+
|        1 |        4 |        7 |
+----------+----------+----------+
|        2 |        5 |        8 |
+----------+----------+----------+
|        3 |        6 |        9 |
+----------+----------+----------+
"""

    def test_add_row_and_column(self):
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )
        table.add_row("This is a single row")
        table.add_column("|rThis is long data|n", "|bThis is even longer data|n")

        expected = """
+----------------------+----------+----------+--------------------------+
| Heading1             | Heading2 | Heading3 |                          |
+~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~~~~~~~~~~~~~~~~~~~+
| 1                    | 4        | 7        | This is long data        |
+----------------------+----------+----------+--------------------------+
| 2                    | 5        | 8        | This is even longer data |
+----------------------+----------+----------+--------------------------+
| 3                    | 6        | 9        |                          |
+----------------------+----------+----------+--------------------------+
| This is a single row |          |          |                          |
+----------------------+----------+----------+--------------------------+
"""
        self._validate(expected, str(table))

    def test_reformat(self):
        table = evtable.EvTable(
            "|yHeading1|n",
            "|gHeading2|n",
            "|rHeading3|n",
            table=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            border="cells",
            align="l",
        )

        # width
        table.reformat(width=50)

        expected = """
+----------------+---------------+---------------+
| Heading1       | Heading2      | Heading3      |
+~~~~~~~~~~~~~~~~+~~~~~~~~~~~~~~~+~~~~~~~~~~~~~~~+
| 1              | 4             | 7             |
+----------------+---------------+---------------+
| 2              | 5             | 8             |
+----------------+---------------+---------------+
| 3              | 6             | 9             |
+----------------+---------------+---------------+
        """
        self._validate(expected, str(table))

        # right-aligned

        table.reformat_column(2, width=30, align="r")

        expected = """
+---------+--------+-----------------------------+
| Heading | Headin |                    Heading3 |
| 1       | g2     |                             |
+~~~~~~~~~+~~~~~~~~+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
| 1       | 4      |                           7 |
+---------+--------+-----------------------------+
| 2       | 5      |                           8 |
+---------+--------+-----------------------------+
| 3       | 6      |                           9 |
+---------+--------+-----------------------------+
        """
        self._validate(expected, str(table))

    def test_multiple_rows(self):
        """
        Adding a lot of rows with `.add_row`.

        """
        table = evtable.EvTable("|yHeading1|n", "|B|[GHeading2|n", "Heading3")
        nlines = 12

        for i in range(nlines):
            table.add_row(
                f"This is col 0, row {i}",
                f"|gThis is col 1, row |w{i}|n|g|n",
                f"This is col 2, row {i}",
            )

        expected = [
            "+-----------------------+-----------------------+-----------------------+",
            "| Heading1              | Heading2              | Heading3              |",
            "+~~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~~~~~~~~~~~~~~~+~~~~~~~~~~~~~~~~~~~~~~~+",
        ]
        for i in range(nlines):
            expected.append(
                f"| This is col 0, row {i:<2} | This is col 1, row {i:<2} | This is col 2, row"
                f" {i:<2} |"
            )
        expected.append(expected[0])
        expected = "\n".join(expected)

        self._validate(expected, str(table))

    def test_direct_evcolumn_adds(self):
        """
        Testing https://github.com/evennia/evennia/issues/2762

        Extra spaces getting added to cell content

        Also testing out adding EvColumns directly to the table kwarg.

        """
        # direct table add
        table = evtable.EvTable(table=[["another"]], fill_char=".", pad_char="#", width=8)

        expected = """
+------+
|#anot#|
|#her.#|
+------+
        """
        self._validate(expected, str(table))

        # add with .add_column
        table = evtable.EvTable(fill_char=".", pad_char="#")
        table.add_column("another", width=8)

        self._validate(expected, str(table))

        # add by passing a column to constructor directly

        colB = evtable.EvColumn("another", width=8)

        table = evtable.EvTable(table=[colB], fill_char=".", pad_char="#")

        self._validate(expected, str(table))

        # more complex table

        colA = evtable.EvColumn("this", "is", "a", "column")  # no width
        colB = evtable.EvColumn("and", "another", "one", "here", width=8)

        table = evtable.EvTable(table=[colA, colB], fill_char=".", pad_char="#")

        expected = """
+--------+-------+
|#this..#|#and..#|
|#is....#|#anoth#|
|#......#|#er...#|
|#a.....#|#one..#|
|#column#|#here.#|
+--------+-------+
        """

        self._validate(expected, str(table))

    def test_width_enforcement(self):
        """
        Testing https://github.com/evennia/evennia/issues/2761

        EvTable enforces width kwarg, expanding the wrong column

        """
        # simple crop
        table = evtable.EvTable(table=[["column"]], width=7, enforce_size=True)
        expected = """
+-----+
| col |
+-----+
        """

        # more advanced table with crop
        self._validate(expected, str(table))

        colA = evtable.EvColumn("it", "is", "a", "column", width=6, enforce_size=True)
        colB = evtable.EvColumn("and", "another", "column", "here")
        table = evtable.EvTable(table=[colA, colB], width=40)

        expected = """
+----+---------------------------------+
| it | and                             |
| is | another                         |
| a  | column                          |
| co | here                            |
+----+---------------------------------+
        """

        self._validate(expected, str(table))

    def test_styling_overrides(self):
        """
        Testing https://github.com/evennia/evennia/issues/2760

        Not being able to override table settings.

        """
        column = evtable.EvColumn("this", "is", "a", "column", fill_char=".")
        table = evtable.EvTable(table=[column])

        expected = """
+--------+
| this.. |
| is.... |
| a..... |
| column |
+--------+
        """

        self._validate(expected, str(table))

    def test_color_transfer(self):
        """
        Testing https://github.com/evennia/evennia/issues/2986

        EvTable swallowing color tags.

        """
        from evennia.utils.ansi import ANSI_CYAN, ANSI_RED

        row1 = "|cAn entire colored row|n"
        row2 = "A single |rred|n word"

        table = evtable.EvTable(table=[[row1, row2]])

        self.assertIn(ANSI_RED, str(table))
        self.assertIn(ANSI_CYAN, str(table))

    @skip("Pending refactor into client-side ansi parsing")
    def test_mxp_links(self):
        """
        Testing https://github.com/evennia/evennia/issues/3082

        EvTable not properly handling mxp links given to it.

        """

        commands1 = [f"|lcsay This is command {inum}|ltcommand {inum}|le" for inum in range(1, 4)]
        commands2 = [f"command {inum}" for inum in range(1, 4)]  # comparison strings, no MXP

        # from evennia import set_trace

        # set_trace()

        cell1 = ansi.strip_mxp(str(evtable.EvCell(f"|lcsay This is command 1|ltcommand 1|le")))
        cell2 = str(evtable.EvCell(f"command 1"))

        print(f"cell1:------------\n{cell1}")
        print(f"cell2:------------\n{cell2}")

        table1a = ansi.strip_mxp(str(evtable.EvTable(*commands1)))
        table1b = str(evtable.EvTable(*commands2))

        table2a = ansi.strip_mxp(str(evtable.EvTable(table=[commands1])))
        table2b = str(evtable.EvTable(table=[commands2]))

        print(f"1a:---------------\n{table1a}")
        print(f"1b:---------------\n{table1b}")
        print(f"2a:---------------\n{table2a}")
        print(f"2b:---------------\n{table2b}")

        self.assertEqual(table1b, table1a)
        self.assertEqual(table2b, table2a)
